/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#ifndef DECODEFLAC_HPP
#define DECODEFLAC_HPP

#include <cstdio>
#include <cstdlib>
#include <cstring>

class DecodeFLAC : public BufferedNode {

}

static FLAC__StreamDecoderWriteStatus write_callback(const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data);
static void metadata_callback(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data);
static void error_callback(const FLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data);

static FLAC__uint64 total_samples = 0;
static unsigned sample_rate = 0;
static unsigned channels = 0;
static unsigned bps = 0;

static FLAC__bool write_little_endian_uint16(FILE *f, FLAC__uint16 x)
{
	return
		fputc(x, f) != EOF &&
		fputc(x >> 8, f) != EOF
	;
}

static FLAC__bool write_little_endian_int16(FILE *f, FLAC__int16 x)
{
	return write_little_endian_uint16(f, (FLAC__uint16)x);
}

static FLAC__bool write_little_endian_uint32(FILE *f, FLAC__uint32 x)
{
	return
		fputc(x, f) != EOF &&
		fputc(x >> 8, f) != EOF &&
		fputc(x >> 16, f) != EOF &&
		fputc(x >> 24, f) != EOF
	;
}

int main(int argc, char *argv[])
{
	FLAC__bool ok = true;
	FLAC__StreamDecoder *decoder = 0;
	FLAC__StreamDecoderInitStatus init_status;
	FILE *fout;

	if(argc != 3) {
		fprintf(stderr, "usage: %s infile.flac outfile.wav\n", argv[0]);
		return 1;
	}

	if((fout = fopen(argv[2], "wb")) == NULL) {
		fprintf(stderr, "ERROR: opening %s for output\n", argv[2]);
		return 1;
	}

	if((decoder = FLAC__stream_decoder_new()) == NULL) {
		fprintf(stderr, "ERROR: allocating decoder\n");
		fclose(fout);
		return 1;
	}

	(void)FLAC__stream_decoder_set_md5_checking(decoder, true);

	init_status = FLAC__stream_decoder_init_file(decoder, argv[1], write_callback, metadata_callback, error_callback, /*client_data=*/fout);
	if(init_status != FLAC__STREAM_DECODER_INIT_STATUS_OK) {
		fprintf(stderr, "ERROR: initializing decoder: %s\n", FLAC__StreamDecoderInitStatusString[init_status]);
		ok = false;
	}

	if(ok) {
		ok = FLAC__stream_decoder_process_until_end_of_stream(decoder);
		fprintf(stderr, "decoding: %s\n", ok? "succeeded" : "FAILED");
		fprintf(stderr, "   state: %s\n", FLAC__StreamDecoderStateString[FLAC__stream_decoder_get_state(decoder)]);
	}

	FLAC__stream_decoder_delete(decoder);
	fclose(fout);

	return 0;
}

FLAC__StreamDecoderWriteStatus write_callback(const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data)
{
	FILE *f = (FILE*)client_data;
	const FLAC__uint32 total_size = (FLAC__uint32)(total_samples * channels * (bps/8));
	size_t i;

	(void)decoder;

	if(total_samples == 0) {
		fprintf(stderr, "ERROR: this example only works for FLAC files that have a total_samples count in STREAMINFO\n");
		return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
	}
	if(channels != 2 || bps != 16) {
		fprintf(stderr, "ERROR: this example only supports 16bit stereo streams\n");
		return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
	}

	/* write WAVE header before we write the first frame */
	if(frame->header.number.sample_number == 0) {
		if(
			fwrite("RIFF", 1, 4, f) < 4 ||
			!write_little_endian_uint32(f, total_size + 36) ||
			fwrite("WAVEfmt ", 1, 8, f) < 8 ||
			!write_little_endian_uint32(f, 16) ||
			!write_little_endian_uint16(f, 1) ||
			!write_little_endian_uint16(f, (FLAC__uint16)channels) ||
			!write_little_endian_uint32(f, sample_rate) ||
			!write_little_endian_uint32(f, sample_rate * channels * (bps/8)) ||
			!write_little_endian_uint16(f, (FLAC__uint16)(channels * (bps/8))) || /* block align */
			!write_little_endian_uint16(f, (FLAC__uint16)bps) ||
			fwrite("data", 1, 4, f) < 4 ||
			!write_little_endian_uint32(f, total_size)
		) {
			fprintf(stderr, "ERROR: write error\n");
			return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
		}
	}

	/* write decoded PCM samples */
	for(i = 0; i < frame->header.blocksize; i++) {
		if(
			!write_little_endian_int16(f, (FLAC__int16)buffer[0][i]) ||  /* left channel */
			!write_little_endian_int16(f, (FLAC__int16)buffer[1][i])     /* right channel */
		) {
			fprintf(stderr, "ERROR: write error\n");
			return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
		}
	}

	return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

void metadata_callback(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data)
{
	(void)decoder, (void)client_data;

	/* print some stats */
	if(metadata->type == FLAC__METADATA_TYPE_STREAMINFO) {
		/* save for later */
		total_samples = metadata->data.stream_info.total_samples;
		sample_rate = metadata->data.stream_info.sample_rate;
		channels = metadata->data.stream_info.channels;
		bps = metadata->data.stream_info.bits_per_sample;

		fprintf(stderr, "sample rate    : %u Hz\n", sample_rate);
		fprintf(stderr, "channels       : %u\n", channels);
		fprintf(stderr, "bits per sample: %u\n", bps);
		fprintf(stderr, "total samples  : %" PRIu64 "\n", total_samples);
	}
}

void error_callback(const FLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data)
{
	(void)decoder, (void)client_data;

	fprintf(stderr, "Got error callback: %s\n", FLAC__StreamDecoderErrorStatusString[status]);
}


















/*
class SubscriberWebSocket : public BufferedNode {
  //  Output IDs | Connected Flags
  int output1ID;  bool boutput1ID;
  int output2ID;  bool boutput2ID;
  int output3ID;  bool boutput3ID;
  int output4ID;  bool boutput4ID;
  int output5ID;  bool boutput5ID;
  int output6ID;  bool boutput6ID;
  int output7ID;  bool boutput7ID;
  int output8ID;  bool boutput8ID;
  int output9ID;  bool boutput9ID;
  int output10ID; bool boutput10ID;
  int output11ID; bool boutput11ID;
  int output12ID; bool boutput12ID;
  int output13ID; bool boutput13ID;
  int output14ID; bool boutput14ID;
  int output15ID; bool boutput15ID;
  int output16ID; bool boutput16ID;

  // Condition in/out
  int input_condition_syncID; bool binput_condition_syncID;
  int output_conditionID;

  // Parameters [b:bool, i:int, f:float, d:double, s:string, o:ObjectRef, cx:Complex<float>, vx:Vector<x>, mx:Matrix<x>, px:Map<int, x>]
  bool   param_b_debug_status;  // Enable debug status         [true:enable, false:disable]
  bool   param_b_debug_sync;    // Enable debug synchronize    [true:enable, false:disable]
  bool   param_b_mode_server;   // Select Server/Client mode   [true:server, false:client]
  bool   param_b_mode_ssl;      // Use SSL connection          [true:enable, false:disable]
  bool   param_b_mode_dosguard; // Use DoS attack guard        [true:enable, false:disable]
  int    param_i_limit_size;    // Receive data limit size for DoS attack guard function
  string param_s_host;          // IP or Hostname
  int    param_i_port;          // Port number
  string param_s_dir;           // Directory suffix
  bool   param_b_timeout;       // Use timeout for client mode [true:enable, false:disable]
  bool   param_i_timeout;       // Timeout[ms] for timeout function
  int    param_i_buffer;        // Buffer size

  int    param_i_length;        // Frame length
  int    param_i_advance;       // Shift size
  int    param_i_channels;      // Number of channels
  int    param_i_sampling_rate; // Sampling rate

  int    param_i_streams;       // Number of target streams
  string param_s_stream1;       // Stream type for each stream
  string param_s_stream2;
  string param_s_stream3;
  string param_s_stream4;
  string param_s_stream5;
  string param_s_stream6;
  string param_s_stream7;
  string param_s_stream8;
  string param_s_stream9;
  string param_s_stream10;
  string param_s_stream11;
  string param_s_stream12;
  string param_s_stream13;
  string param_s_stream14;
  string param_s_stream15;
  string param_s_stream16;

  bool   param_b_overlap1;  bool boutput1_overlap;  bool boutput1_disoverlap;  bool boutput1_autooverlap;
  bool   param_b_overlap2;  bool boutput2_overlap;  bool boutput2_disoverlap;  bool boutput2_autooverlap;
  bool   param_b_overlap3;  bool boutput3_overlap;  bool boutput3_disoverlap;  bool boutput3_autooverlap;
  bool   param_b_overlap4;  bool boutput4_overlap;  bool boutput4_disoverlap;  bool boutput4_autooverlap;
  bool   param_b_overlap5;  bool boutput5_overlap;  bool boutput5_disoverlap;  bool boutput5_autooverlap;
  bool   param_b_overlap6;  bool boutput6_overlap;  bool boutput6_disoverlap;  bool boutput6_autooverlap;
  bool   param_b_overlap7;  bool boutput7_overlap;  bool boutput7_disoverlap;  bool boutput7_autooverlap;
  bool   param_b_overlap8;  bool boutput8_overlap;  bool boutput8_disoverlap;  bool boutput8_autooverlap;
  bool   param_b_overlap9;  bool boutput9_overlap;  bool boutput9_disoverlap;  bool boutput9_autooverlap;
  bool   param_b_overlap10; bool boutput10_overlap; bool boutput10_disoverlap; bool boutput10_autooverlap;
  bool   param_b_overlap11; bool boutput11_overlap; bool boutput11_disoverlap; bool boutput11_autooverlap;
  bool   param_b_overlap12; bool boutput12_overlap; bool boutput12_disoverlap; bool boutput12_autooverlap;
  bool   param_b_overlap13; bool boutput13_overlap; bool boutput13_disoverlap; bool boutput13_autooverlap;
  bool   param_b_overlap14; bool boutput14_overlap; bool boutput14_disoverlap; bool boutput14_autooverlap;
  bool   param_b_overlap15; bool boutput15_overlap; bool boutput15_disoverlap; bool boutput15_autooverlap;
  bool   param_b_overlap16; bool boutput16_overlap; bool boutput16_disoverlap; bool boutput16_autooverlap;

  int c_sock;         // ファイルディスクリプタ
  int sock;
  int yes;
  struct sockaddr_in addr;
  struct sockaddr_in client;
  short int *buffer;  // バッファ
  int buf_size;       // バッファサイズ
  
  int isConnected;

  public:
  // Constructor
  SubscriberWebSocket(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params),
    output1ID(-1),  output2ID(-1),  output3ID(-1),  output4ID(-1),
    output5ID(-1),  output6ID(-1),  output7ID(-1),  output8ID(-1),
    output9ID(-1),  output10ID(-1), output11ID(-1), output12ID(-1),
    output13ID(-1), output14ID(-1), output15ID(-1), output16ID(-1),
    boutput1ID(false),  boutput2ID(false),  boutput3ID(false),  boutput4ID(false),
    boutput5ID(false),  boutput6ID(false),  boutput7ID(false),  boutput8ID(false),
    boutput9ID(false),  boutput10ID(false), boutput11ID(false), boutput12ID(false),
    boutput13ID(false), boutput14ID(false), boutput15ID(false), boutput16ID(false),
    boutput1_overlap(false),  boutput2_overlap(false),  boutput3_overlap(false),  boutput4_overlap(false),
    boutput5_overlap(false),  boutput6_overlap(false),  boutput7_overlap(false),  boutput8_overlap(false),
    boutput9_overlap(false),  boutput10_overlap(false), boutput11_overlap(false), boutput12_overlap(false),
    boutput13_overlap(false), boutput14_overlap(false), boutput15_overlap(false), boutput16_overlap(false),
    boutput1_disoverlap(false),  boutput2_disoverlap(false),  boutput3_disoverlap(false),  boutput4_disoverlap(false),
    boutput5_disoverlap(false),  boutput6_disoverlap(false),  boutput7_disoverlap(false),  boutput8_disoverlap(false),
    boutput9_disoverlap(false),  boutput10_disoverlap(false), boutput11_disoverlap(false), boutput12_disoverlap(false),
    boutput13_disoverlap(false), boutput14_disoverlap(false), boutput15_disoverlap(false), boutput16_disoverlap(false),
    boutput1_autooverlap(false),  boutput2_autooverlap(false),  boutput3_autooverlap(false),  boutput4_autooverlap(false),
    boutput5_autooverlap(false),  boutput6_autooverlap(false),  boutput7_autooverlap(false),  boutput8_autooverlap(false),
    boutput9_autooverlap(false),  boutput10_autooverlap(false), boutput11_autooverlap(false), boutput12_autooverlap(false),
    boutput13_autooverlap(false), boutput14_autooverlap(false), boutput15_autooverlap(false), boutput16_autooverlap(false),
    input_condition_syncID(-1), binput_condition_syncID(false),
    num_channels(0)
  {
  }
    
  ~SubscriberWebSocket()
  {
  }

  // Initialize
  virtual void initialize() {
  }

  // dynamic input-port translation
  virtual int translateInput(string inputName) {
  }

  // dynamic output-port translation
  virtual int translateOutput (string outputName)
  {
    return addOutput(outputName);
  }

  void calculate(int output_id, int count, Buffer &out){
  }

  ///////////////////////////////////////////////////////////////
  // WebSocket server start
  ///////////////////////////////////////////////////////////////
  int startServer(int port, bool use_limit, int limit){
    int status_code = 0;
    return status_code;
  }

  ///////////////////////////////////////////////////////////////
  // WebSocket client start
  ///////////////////////////////////////////////////////////////
  int startClient(int port, bool use_limit, int limit){
    int status_code = 0;
    return status_code; // 1:success, 0:throw, -1:exception
  }
};
*/
#endif // DECODEFLAC_HPP

