/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#ifndef ENCODEFLAC_HPP
#define ENCODEFLAC_HPP

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include "share/compat.h"
#include "FLAC/metadata.h"
#include "FLAC/stream_encoder.h"

static void progress_callback(const FLAC__StreamEncoder *encoder, FLAC__uint64 bytes_written, FLAC__uint64 samples_written, unsigned frames_written, unsigned total_frames_estimate, void *client_data);

#define READSIZE 1024

static unsigned total_samples = 0; /* can use a 32-bit number due to WAVE size limitations */
static FLAC__byte buffer[READSIZE/*samples*/ * 2/*bytes_per_sample*/ * 2/*channels*/]; /* we read the WAVE data into here */
static FLAC__int32 pcm[READSIZE/*samples*/ * 2/*channels*/];

int main(int argc, char *argv[])
{
	FLAC__bool ok = true;
	FLAC__StreamEncoder *encoder = 0;
	FLAC__StreamEncoderInitStatus init_status;
	FLAC__StreamMetadata *metadata[2];
	FLAC__StreamMetadata_VorbisComment_Entry entry;
	FILE *fin;
	unsigned sample_rate = 0;
	unsigned channels = 0;
	unsigned bps = 0;

	if(argc != 3) {
		fprintf(stderr, "usage: %s infile.wav outfile.flac\n", argv[0]);
		return 1;
	}

	if((fin = fopen(argv[1], "rb")) == NULL) {
		fprintf(stderr, "ERROR: opening %s for output\n", argv[1]);
		return 1;
	}

	/* read wav header and validate it */
	if(
		fread(buffer, 1, 44, fin) != 44 ||
		memcmp(buffer, "RIFF", 4) ||
		memcmp(buffer+8, "WAVEfmt \020\000\000\000\001\000\002\000", 16) ||
		memcmp(buffer+32, "\004\000\020\000data", 8)
	) {
		fprintf(stderr, "ERROR: invalid/unsupported WAVE file, only 16bps stereo WAVE in canonical form allowed\n");
		fclose(fin);
		return 1;
	}
	sample_rate = ((((((unsigned)buffer[27] << 8) | buffer[26]) << 8) | buffer[25]) << 8) | buffer[24];
	channels = 2;
	bps = 16;
	total_samples = (((((((unsigned)buffer[43] << 8) | buffer[42]) << 8) | buffer[41]) << 8) | buffer[40]) / 4;

	/* allocate the encoder */
	if((encoder = FLAC__stream_encoder_new()) == NULL) {
		fprintf(stderr, "ERROR: allocating encoder\n");
		fclose(fin);
		return 1;
	}

	ok &= FLAC__stream_encoder_set_verify(encoder, true);
	ok &= FLAC__stream_encoder_set_compression_level(encoder, 5);
	ok &= FLAC__stream_encoder_set_channels(encoder, channels);
	ok &= FLAC__stream_encoder_set_bits_per_sample(encoder, bps);
	ok &= FLAC__stream_encoder_set_sample_rate(encoder, sample_rate);
	ok &= FLAC__stream_encoder_set_total_samples_estimate(encoder, total_samples);

	/* now add some metadata; we'll add some tags and a padding block */
	if(ok) {
		if(
			(metadata[0] = FLAC__metadata_object_new(FLAC__METADATA_TYPE_VORBIS_COMMENT)) == NULL ||
			(metadata[1] = FLAC__metadata_object_new(FLAC__METADATA_TYPE_PADDING)) == NULL ||
			/* there are many tag (vorbiscomment) functions but these are convenient for this particular use: */
			!FLAC__metadata_object_vorbiscomment_entry_from_name_value_pair(&entry, "ARTIST", "Some Artist") ||
			!FLAC__metadata_object_vorbiscomment_append_comment(metadata[0], entry, /*copy=*/false) || /* copy=false: let metadata object take control of entry's allocated string */
			!FLAC__metadata_object_vorbiscomment_entry_from_name_value_pair(&entry, "YEAR", "1984") ||
			!FLAC__metadata_object_vorbiscomment_append_comment(metadata[0], entry, /*copy=*/false)
		) {
			fprintf(stderr, "ERROR: out of memory or tag error\n");
			ok = false;
		}

		metadata[1]->length = 1234; /* set the padding length */

		ok = FLAC__stream_encoder_set_metadata(encoder, metadata, 2);
	}

	/* initialize encoder */
	if(ok) {
		init_status = FLAC__stream_encoder_init_file(encoder, argv[2], progress_callback, /*client_data=*/NULL);
		if(init_status != FLAC__STREAM_ENCODER_INIT_STATUS_OK) {
			fprintf(stderr, "ERROR: initializing encoder: %s\n", FLAC__StreamEncoderInitStatusString[init_status]);
			ok = false;
		}
	}

	/* read blocks of samples from WAVE file and feed to encoder */
	if(ok) {
		size_t left = (size_t)total_samples;
		while(ok && left) {
			size_t need = (left>READSIZE? (size_t)READSIZE : (size_t)left);
			if(fread(buffer, channels*(bps/8), need, fin) != need) {
				fprintf(stderr, "ERROR: reading from WAVE file\n");
				ok = false;
			}
			else {
				/* convert the packed little-endian 16-bit PCM samples from WAVE into an interleaved FLAC__int32 buffer for libFLAC */
				size_t i;
				for(i = 0; i < need*channels; i++) {
					/* inefficient but simple and works on big- or little-endian machines */
					pcm[i] = (FLAC__int32)(((FLAC__int16)(FLAC__int8)buffer[2*i+1] << 8) | (FLAC__int16)buffer[2*i]);
				}
				/* feed samples to encoder */
				ok = FLAC__stream_encoder_process_interleaved(encoder, pcm, need);
			}
			left -= need;
		}
	}

	ok &= FLAC__stream_encoder_finish(encoder);

	fprintf(stderr, "encoding: %s\n", ok? "succeeded" : "FAILED");
	fprintf(stderr, "   state: %s\n", FLAC__StreamEncoderStateString[FLAC__stream_encoder_get_state(encoder)]);

	/* now that encoding is finished, the metadata can be freed */
	FLAC__metadata_object_delete(metadata[0]);
	FLAC__metadata_object_delete(metadata[1]);

	FLAC__stream_encoder_delete(encoder);
	fclose(fin);

	return 0;
}

void progress_callback(const FLAC__StreamEncoder *encoder, FLAC__uint64 bytes_written, FLAC__uint64 samples_written, unsigned frames_written, unsigned total_frames_estimate, void *client_data)
{
	(void)encoder, (void)client_data;

	fprintf(stderr, "wrote %" PRIu64 " bytes, %" PRIu64 "/%u samples, %u/%u frames\n", bytes_written, samples_written, total_samples, frames_written, total_frames_estimate);
}

















































/*
//#ifdef ENABLE_WEBSOCKET
#include "roles/server.hpp"
#include "roles/client.hpp"
#include "websocketpp.hpp"

#include "WebSocket.hpp"
#include "PublisherWebSocket.hpp"

//#include "HarkParamsObjects.h"

#include <boost/algorithm/string/replace.hpp>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/thread.hpp>

using namespace FD;
using namespace std;

using namespace websocketconnector;
using boost::asio::ip::tcp;
using websocketpp::server;
using websocketpp::client;


enum sss {
  STREAM_IS_DATA  = 0x01000000,
  STREAM_IS_PARAM = 0x02000000,
  STREAM_IS_RSV03 = 0x03000000,
  STREAM_IS_RSV04 = 0x04000000,
  STREAM_IS_RSV05 = 0x05000000,
  STREAM_IS_RSV06 = 0x06000000,
  STREAM_IS_RSV07 = 0x07000000,
  STREAM_IS_RSV08 = 0x08000000,
  STREAM_IS_RSV09 = 0x09000000,
  STREAM_IS_RSV10 = 0x0A000000,
  STREAM_IS_RSV11 = 0x0B000000,
  STREAM_IS_RSV12 = 0x0C000000,
  STREAM_IS_RSV13 = 0x0D000000,
  STREAM_IS_RSV14 = 0x0E000000,
  STREAM_IS_RSV15 = 0x0F000000,

  // If STREAM_IS_DATA
  DSTREAM_IS_THROUGH = 0x01000000, // STREAM??_OVERLAPCONTROL = Through
  DSTREAM_IS_ADDFLAG = 0x02000000, // STREAM??_OVERLAPCONTROL = AddFlag
  DSTREAM_IS_OVERLAP = 0x03000000, // STREAM??_OVERLAPCONTROL = Overlap
  DSTREAM_IS_DISOLAP = 0x04000000, // STREAM??_OVERLAPCONTROL = DisOverlap
  DSTREAM_IS_RSV05   = 0x05000000,
  DSTREAM_IS_RSV06   = 0x06000000,
  DSTREAM_IS_RSV07   = 0x07000000,
  DSTREAM_IS_RSV08   = 0x08000000,
  DSTREAM_IS_RSV09   = 0x09000000,
  DSTREAM_IS_RSV10   = 0x0A000000,
  DSTREAM_IS_RSV11   = 0x0B000000,
  DSTREAM_IS_RSV12   = 0x0C000000,
  DSTREAM_IS_RSV13   = 0x0D000000,
  DSTREAM_IS_RSV14   = 0x0E000000,
  DSTREAM_IS_RSV15   = 0x0F000000,

  // If STREAM_IS_PARAM
  PSTREAM_IS_DATA  = 0x01000000,
  PSTREAM_IS_PARAM = 0x02000000,
  PSTREAM_IS_RSV03 = 0x03000000,
  PSTREAM_IS_RSV04 = 0x04000000,
  PSTREAM_IS_RSV05 = 0x05000000,
  PSTREAM_IS_RSV06 = 0x06000000,
  PSTREAM_IS_RSV07 = 0x07000000,
  PSTREAM_IS_RSV08 = 0x08000000,
  PSTREAM_IS_RSV09 = 0x09000000,
  PSTREAM_IS_RSV10 = 0x0A000000,
  PSTREAM_IS_RSV11 = 0x0B000000,
  PSTREAM_IS_RSV12 = 0x0C000000,
  PSTREAM_IS_RSV13 = 0x0D000000,
  PSTREAM_IS_RSV14 = 0x0E000000,
  PSTREAM_IS_RSV15 = 0x0F000000,

  STREAM_IS_DATA  = 0x10000000, // STREAM_SELECTION_?? = DataStream
  STREAM_IS_PARAM = 0x20000000, // STREAM_SELECTION_?? = ParamStream
  STREAM_IS_RSV03 = 0x30000000,
  STREAM_IS_RSV04 = 0x40000000,
  STREAM_IS_RSV05 = 0x50000000,
  STREAM_IS_RSV06 = 0x60000000,
  STREAM_IS_RSV07 = 0x70000000,
  STREAM_IS_RSV08 = 0x80000000,
  DUMMY           = 0xFFFFFFFF
};

class PublisherWebSocket : public BufferedNode {
  //  Input IDs | Connected Flags
  int input1ID;  bool binput1ID;  bool binput1proc;
  int input2ID;  bool binput2ID;  bool binput2proc;
  int input3ID;  bool binput3ID;  bool binput3proc;
  int input4ID;  bool binput4ID;  bool binput4proc;
  int input5ID;  bool binput5ID;  bool binput5proc;
  int input6ID;  bool binput6ID;  bool binput6proc;
  int input7ID;  bool binput7ID;  bool binput7proc;
  int input8ID;  bool binput8ID;  bool binput8proc;
  int input9ID;  bool binput9ID;  bool binput9proc;
  int input10ID; bool binput10ID; bool binput10proc;
  int input11ID; bool binput11ID; bool binput11proc;
  int input12ID; bool binput12ID; bool binput12proc;
  int input13ID; bool binput13ID; bool binput13proc;
  int input14ID; bool binput14ID; bool binput14proc;
  int input15ID; bool binput15ID; bool binput15proc;
  int input16ID; bool binput16ID; bool binput16proc;

  // Condition in/out
  int input_condition_syncID; bool binput_condition_syncID;
  int output_conditionID;

  // Parameters [b:bool, i:int, f:float, d:double, s:string, o:ObjectRef, cx:Complex<float>, vx:Vector<x>, mx:Matrix<x>, px:Map<int, x>]
  bool   param_b_debug_status;  // Enable debug status         [true:enable, false:disable]
  bool   param_b_debug_sync;    // Enable debug synchronize    [true:enable, false:disable]
  bool   param_b_mode_server;   // Select Server/Client mode   [true:server, false:client]
  bool   param_b_mode_ssl;      // Use SSL connection          [true:enable, false:disable]
  bool   param_b_mode_dosguard; // Use DoS attack guard        [true:enable, false:disable]
  int    param_i_limit_size;    // Receive data limit size for DoS attack guard function
  string param_s_host;          // IP or Hostname
  int    param_i_port;          // Port number
  string param_s_dir;           // Directory suffix
  bool   param_b_timeout;       // Use timeout for client mode [true:enable, false:disable]
  bool   param_i_timeout;       // Timeout[ms] for timeout function
  int    param_i_buffer;        // Buffer size

  int    param_i_length;        // Frame length
  int    param_i_advance;       // Shift size
  int    param_i_channels;      // Number of channels
  int    param_i_sampling_rate; // Sampling rate

  int    param_i_streams;       // Number of target streams
  string param_s_stream1;       // Stream type for each stream
  string param_s_stream2;
  string param_s_stream3;
  string param_s_stream4;
  string param_s_stream5;
  string param_s_stream6;
  string param_s_stream7;
  string param_s_stream8;
  string param_s_stream9;
  string param_s_stream10;
  string param_s_stream11;
  string param_s_stream12;
  string param_s_stream13;
  string param_s_stream14;
  string param_s_stream15;
  string param_s_stream16;

  // for other machine environment compatibility.
  int header_magic;      // Magic Code, If you read this magic code by int on 32bit system
  int header_padding;    // for 64bit system guard bits, Maybe 64bit system is if you are reading along with the padding when you read the magic code.
  int header_reserved1;
  int header_version;
  int header_sz_char;
  int header_sz_int;
  int header_sz_float;
  int header_sz_double;
  int header_reserved2;
  string header_string("");

  ofstream output_file_stream;
  bool bstream_opend;

  public:
  // Constructor
  PublisherWebSocket(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params),
    input1ID(-1),        input2ID(-1),        input3ID(-1),        input4ID(-1),
    input5ID(-1),        input6ID(-1),        input7ID(-1),        input8ID(-1),
    input9ID(-1),        input10ID(-1),       input11ID(-1),       input12ID(-1),
    input13ID(-1),       input14ID(-1),       input15ID(-1),       input16ID(-1),
    binput1ID(false),    binput2ID(false),    binput3ID(false),    binput4ID(false),
    binput5ID(false),    binput6ID(false),    binput7ID(false),    binput8ID(false),
    binput9ID(false),    binput10ID(false),   binput11ID(false),   binput12ID(false),
    binput13ID(false),   binput14ID(false),   binput15ID(false),   binput16ID(false),
    binput1proc(false),  binput2proc(false),  binput3proc(false),  binput4proc(false),
    binput5proc(false),  binput6proc(false),  binput7proc(false),  binput8proc(false),
    binput9proc(false),  binput10proc(false), binput11proc(false), binput12proc(false),
    binput13proc(false), binput14proc(false), binput15proc(false), binput16proc(false),
    input_condition_syncID(-1), binput_condition_syncID(false),
    bstream_opend(false),
    num_channels(0)
  {
  }

  ~PublisherWebSocket()
  {
  }


  // dynamic input-port translation
  virtual int translateInput(string inputName) {
  }

  // Initialize
  virtual void initialize() {
  }

  // Implementation
  void calculate(int output_id, int count, Buffer &out){
  }

  ///////////////////////////////////////////////////////////////
  // WebSocket server start
  ///////////////////////////////////////////////////////////////
  int startServer(int port, bool use_limit, int limit){
    int status_code = 0;
    return status_code;
  }


  ///////////////////////////////////////////////////////////////
  // WebSocket client start
  ///////////////////////////////////////////////////////////////
  int startClient(int port, bool use_limit, int limit){
    int status_code = 0;
    return status_code; // 1:success, 0:throw, -1:exception
  }

};

*/
#endif // ENCODEFLAC_HPP

