/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <sstream>
#include <iomanip>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Source.h>
#include <Map.h>
#include <TimeStamp.h>
#include <Stream.h>
#include <math.h>

#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>

//#ifdef ENABLE_WEBSOCKET

#include "WebSocket.hpp"
#include <boost/shared_ptr.hpp>

//#include "websocketpp/roles/server.hpp"
#include "websocketpp/roles/client.hpp"
#include "websocketpp/websocketpp.hpp"

#include <boost/algorithm/string/replace.hpp>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/thread.hpp>
//#include "PublisherWebSocket.hpp"

#include <map>
#include <string>
#include <queue>

//#include "HarkParamsObjects.h"

using namespace FD;
using namespace std;

using boost::asio::ip::tcp;
using websocketpp::server;
using websocketpp::client;

//using namespace websocketconnector;
//using namespace websocketconnector::ws_client_handler;
//using namespace websocketconnector::ws_server_handler;

class PublisherWebSocket;

DECLARE_NODE(PublisherWebSocket)
/*Node
 *
 * @name PublisherWebSocket
 * @category HARKD:AudioIO
 * @description Many stream data (Ex: Multi-channel sound, FFT, Source ID, etc...) from the WebSocket network.
 *
 * @parameter_name ENABLE_EXPORT
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Encoded flac data export to file.
 *
 * @parameter_name EXPORT_FILENAME
 * @parameter_type string
 * @parameter_value EXPORT_DEBUG_FILE
 * @parameter_valid ENABLE_EXPORT_DEBUG=true
 * @parameter_description Export filename.
 *
 * @parameter_name ENABLE_STATUS_DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description flac encoder status print debug message of this module in case of true.
 *
 * @parameter_name OUTPUT_UNIT
 * @parameter_type string
 * @parameter_value SOURCE
 * @parameter_list FRAME:SOURCE
 * @parameter_description Output unit.
 *
 * @parameter_name UNIT_SIZE
 * @parameter_type int
 * @parameter_value 100
 * @parameter_description flac frame size for OUTPUT_UNIT is FRAME.
 * 
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The frame length of each channel (in samples) [default: 512].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length beween adjacent frames (in samples)[default: 160].
 *
 * @parameter_name FLAC_CHANNEL
 * @parameter_type int
 * @parameter_value 8
 * @parameter_list 1:2:3:4:5:6:7:8
 * @parameter_description The number of microphone. Support channels listed up from FLAC library API's specification.
 *
 * @parameter_name FLAC_SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_list 88200:176400:192000:8000:16000:22050:24000:32000:44100:48000:96000
 * @parameter_description Sampling rate (Hz) [default: 16000]. Support sampling-rate listed up from FLAC library API's specification.
 *
 * @parameter_name FLAC_BIT_PER_SAMPLE
 * @parameter_type int
 * @parameter_value 24
 * @parameter_list 8:12:16:20:24
 * @parameter_description Selector of FLAC's bit per sample. FLAC default is 16[bits], but HARK default is 24[bits] because HARK is using by many 24 bits microphone.
 *
 * @parameter_name FLAC_ENCODE_LEVEL
 * @parameter_type int
 * @parameter_value 5
 * @parameter_list 0:1:2:3:4:5:6:7:8:15
 * @parameter_description FLAC compression level 0-8, reserved level 9-14 and manual setting mode (=15). Recommended strongly that leave the 5.
 *
 * @parameter_name FLAC_DO_MID_SIDE_STEREO
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Set to true to enable mid-side encoding on stereo input. The number of channels must be 2 for this to have any effect. Set to false to use only independent channel coding. Default value as same as FLAC Compression level 5.
 *
 * @parameter_name FLAC_LOOSE_MID_SIDE_STEREO
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Set to true to enable adaptive switching between mid-side and left-right encoding on stereo input. Set to false to use exhaustive searching. Setting this to true requires FLAC_DO_MID_SIDE_STEREO to also be set to true in order to have any effect. Default value as same as FLAC Compression level 5.
 *
 * @parameter_name FLAC_APODIZATION
 * @parameter_type string
 * @parameter_value tukey(0.5)
 * @parameter_list bartlett:bartlett_hann:blackman:blackman_harris_4term_92db:connes:flattop:gauss(0.1):gauss(0.2):gauss(0.3):gauss(0.4):gauss(0.5):hamming:hann:kaiser_bessel:nuttall:rectangle:triangle:tukey(0.125):tukey(0.25):tukey(0.5):welch
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Sets the apodization function(s) the encoder will use when windowing audio data for LPC analysis. Note: tukey(0.0)==rectangle, tukey(1.0)==hann. Default value as same as FLAC Compression level 5.
 *
 * @parameter_name FLAC_MAX_LPC_ORDER
 * @parameter_type int
 * @parameter_value 8
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Set the maximum LPC order, or 0 to use only the fixed predictors. Default value as same as FLAC Compression level 5.
 *
 * @parameter_name FLAC_QLP_COEFF_PRECISION
 * @parameter_type int
 * @parameter_value 0
 * @parameter_list 0:1:2:3:4:5:6:7:8:9:10:11:12:13:14:15:16:17:18:19:20:21:22:23
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Set the precision, in bits. of the quantized linear predictor coefficients. If set to 0, to let the encoder select it based on the blocksize. In the current implementation, FLAC_QLP_COEFF_PRECISION + FLAC_BIT_PER_SAMPLE must be less than 32. Default value as same as FLAC Compression level 5.
 *
 * @parameter_name FLAC_DO_QLP_COEFF_PREC_SEARCH
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Set to false to use only the specified quantized linear predictor coefficient precision, or true to search neighboring precision values and use the best one. Default value as same as FLAC Compression level 5.
 *
 * @parameter_name FLAC_DO_ESCAPE_CODING
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list false
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Deprecated. Setting this value has no effect.
 *
 * @parameter_name FLAC_DO_EXHAUSTIVE_MODEL_SEARCH
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Set to false to let the encoder estimate the best model order based on the residual signal energy, or true to force the encoder to evaluate all order models and select the best. However, true is very heavy. Default value as same as FLAC Compression level 5.
 *
 * @parameter_name FLAC_MIN_RESIDUAL_PARTITION_ORDER
 * @parameter_type int
 * @parameter_value 0
 * @parameter_list 0:1:2:3:4:5:6:7:8:9:10:11:12:13:14:15:16
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Set the minimum partition order to search when coding the residual. The partition order determines the context size in the residual. The context size will be approximately "blocksize / (2 ^ order)". Default value as same as FLAC Compression level 5.
 *
 * @parameter_name FLAC_MAX_RESIDUAL_PARTITION_ORDER
 * @parameter_type int
 * @parameter_value 5
 * @parameter_list 0:1:2:3:4:5:6:7:8:9:10:11:12:13:14:15:16
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Set the maximum partition order to search when coding the residual. The partition order determines the context size in the residual. The context size will be approximately "blocksize / (2 ^ order)". Default value as same as FLAC Compression level 5.
 *
 * @parameter_name FLAC_RICE_PARAMETER_SEARCH_DIST
 * @parameter_type int
 * @parameter_value 0
 * @parameter_list 0
 * @parameter_valid FLAC_ENCODE_LEVEL=15
 * @parameter_description Deprecated. Setting this value has no effect.
 *
 * @parameter_name STREAM0_PARAM_CONTROLTYPE
 * @parameter_type string
 * @parameter_value HoldAndUpdate
 * @parameter_list HoldAndUpdate:HoldAndSmoothlyUpdate:LedInitial:ZeroInit:OneInit:LikeGeyser
 * @parameter_valid STREAM_SELECTION_0=ParamStream
 * @parameter_description This item is void, because this function is pending. Future function overview: Please select Input value control type.
 *
 * @parameter_name SOURCE_HANDLING
 * @parameter_type string
 * @parameter_value FirstFrame
 * @parameter_list FirstFrame:LastFrame:MostStayed:Average
 * @parameter_description Selector of soure handling type. if you want to result like the "SpeechRecognitionClient and Julius_mft" 's source handling, please set to FirstFrame. Maybe best case MostStayed. Average isn't recommended, because likely to be arbitrary coordinates calculated mathematically that does not exist in the transfer function is high.
 *
 * @parameter_name MATRIX_TO_MAP_TYPE
 * @parameter_type string
 * @parameter_value MultiChannel
 * @parameter_list MultiChannel:EachChannel
 * @parameter_description Selector of output flac type. if you want to multi-channel flac on single stream, please set to multi-channel. if you want to each-channel flac on multi stream, please set to each-channel. This setting is valid only when the Matrix was entered. Other input is determined automatically.
 *
 * @input_name IN_SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @input_name IN_PCM
 * @input_type any
 * @input_description The multi-channel PCM data (type is Matrix) or the PCM data with ID specified by "Source" (type is Map).
 *
 * @output_name OUT_SOURCES
 * @output_type bool
 * @output_description The source location with ID. the output content determine based on IN_SOURCE and SOURCE_HANDLING.
 *
 * @output_name OUT_FLAC
 * @output_type Map<int, ObjectRef>
 * @output_description The multi-channel FLAC data with ID specified by "constant 0" (if input was Matrix, and MATRIX_TO_MAP_TYPE set to MultiChannel). Or each-channel FLAC data with ID specified by "Channel" (if input was Matrix, and MATRIX_TO_MAP_TYPE set to EachChannel). Or the FLAC data with ID specified by "Source" (if input was Map).
 *

 END*/

/*
template <typename T>
string to_bstring(T arg){
  char *buf = (char*)malloc(sizeof(int) + sizeof(T));
  int arg_size = sizeof(T);
  if(buf == NULL){ return ""; }
  for(int i=0; i<sizeof(int); i++){ buf[i] = }
  for(int i=0; i<sizeof(T); i++){ buf[i] = (reinterpret_cast<char*>(&arg))[i]; }
  string ret_str(buf);
  free(buf);
  return ret_str;
}

template <typename T>
T from_bstring(string arg, int size){
  char *buf = (char*)malloc(sizeof(T));
  if(buf == NULL){ return ""; }
  for(int i=0; i<sizeof(T); i++){ buf[i] = (reinterpret_cast<char*>(&arg))[i]; }
  string ret_str(buf);
  free(buf);
  return ret_str;
}
*/



class EncodeFLAC : public BufferedNode {
  //  Input IDs | Connected Flags
  int inputID[16]; bool binputID[16]; bool binputProc[16];
  int input_transmission_controlID; bool binput_transmission_controlID;

  // Condition in/out
  int input_condition_syncID; bool binput_condition_syncID;
  int output_conditionID;
  int output_dummyID;

  // Parameters [b:bool, i:int, f:float, d:double, s:string, o:ObjectRef, cx:Complex<float>, vx:Vector<x>, mx:Matrix<x>, px:Map<int, x>]
  bool   param_b_export;
  string param_s_export_filename;
  bool   param_b_mode_server;   // Select Server/Client mode   [true:server, false:client]
  bool   param_b_mode_ssl;      // Use SSL connection          [true:enable, false:disable]
  bool   param_b_mode_dosguard; // Use DoS attack guard        [true:enable, false:disable]
  int    param_i_limit_size;    // Receive data limit size for DoS attack guard function
  string param_s_host;          // IP or Hostname
  int    param_i_port;          // Port number
  string param_s_dir;           // Directory suffix
  bool   param_b_timeout;       // Use timeout for client mode [true:enable, false:disable]
  bool   param_i_timeout;       // Timeout[ms] for timeout function
  int    param_i_buffer;        // Buffer size

  int    param_i_length;        // Frame length
  int    param_i_advance;       // Shift size
  int    param_i_channels;      // Number of channels
  int    param_i_sampling_rate; // Sampling rate

  int    param_i_streams;       // Number of target streams
  Vector<int> param_vi_skip_frame_trigger_streams;
  bool bskipcheckstream[16]; bool bisnodata[16]; bool bprevframestatus; bool benableskip; // Skip trigger stream?, No data stream?, Previous frame is skipped?, Enable frame skip function?

  int param_i_stream_type[16];     // Stream type for each stream [DataStream:0(default), ParamStream:1]
  int param_i_map_type[16];        // Map type for each stream [MapIsFloat:0(default), MapIsComplexFloat:1]
  int param_i_overlap_control[16]; // Overlap control [Through:0(default), AddFlag:1, Overlap:2, DisOverlap:3]
  int param_i_param_type[16];      // Parameter type [Bool:0(default), Int:1, Float:2, Double:3, String:4, Direct:5, ParamsLocalizeMUSIC:6, ParamsSourceTracker:7, ParamsHRLE:8]
  int param_i_param_control[16];   // Parameter update timing control [HoldAndUpdate:0(default), HoldAndSmoothlyUpdate:1, LedInitial:2, ZeroInit:3, OneInit:4, LikeGeyser:5]

  // for other machine environment compatibility.
  int header_magic;      // Magic Code, If you read this magic code by int on 32bit system
  int header_padding;    // for 64bit system guard bits, Maybe 64bit system is if you are reading along with the padding when you read the magic code.
  int header_reserved1;
  int header_version;
  int header_sz_char;
  int header_sz_int;
  int header_sz_float;
  int header_sz_double;
  int header_reserved2;
  string header_string;

/*
  int c_sock;         // ファイルディスクリプタ
  int sock;
  int yes;
  struct sockaddr_in addr;
  struct sockaddr_in client;
  short int *buffer;  // バッファ
  int buf_size;       // バッファサイズ
  
  int isConnected;
*/
  ofstream output_file_stream;
  bool bstream_opend;

  bool socket_condition;

  int dummy;

  // for client
  //boost::shared_ptr<websocketconnector::ws_client_handler_ptr> m_cli_handler;
  websocketconnector::ws_client_handler_ptr m_cli_handler;
  websocketpp::client *m_cli_endpoint;
  websocketpp::client::connection_ptr m_connect_ptr;

  // for server
  websocketconnector::ws_server_handler_ptr m_srv_handler;
  websocketpp::server *m_srv_endpoint;

  // for multi thread function ... pending
  boost::shared_ptr<boost::thread> m_thread_hark;




  // for each source
  typedef RCPtr<Vector<float> > FVectorPtr;
  typedef RCPtr<Matrix<float> > FMatrixPtr;

  struct flacs
  
  int param_i_bits;
  int bits;
  int bytes_per_sample;

  public:
  // Constructor
  EncodeFLAC(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params),
    input_transmission_controlID(-1), binput_transmission_controlID(false),
    input_condition_syncID(-1), binput_condition_syncID(false),
    bstream_opend(false),
    socket_condition(true),
    dummy(0)
  {
    
    param_s_basename = object_cast<String>(parameters.get("BASENAME"));
    param_i_advance = dereference_cast<int>(parameters.get("ADVANCE"));
    param_i_bits = dereference_cast<int>(parameters.get("FLAC_BIT_PER_SAMPLE"));
    
    switch(param_i_bits){
    case 8:  // unsupported now => change to 16bit mode.
    case 12: // unsupported now => change to 16bit mode.
    case 16:
      bits = 16;
      bytes_per_sample = 2;
      break;
    case 20: // unsupported now => change to 24bit mode.
    case 24: // default
    case 32: // !!invalid value!! 32bit data cannot support on FLAC specification. change to 24bit mode.
    default:
      bits = 24;
      bytes_per_sample = 3;
      break;
    }
    
    
    
    
    
    
    
    
    // Initialize
    for(int stream=0; stream<16; stream++){
      inputID[stream]                 = -1;
      binputID[stream]                = false;
      binputProc[stream]              = false;
      bskipcheckstream[stream]        = false;
      bisnodata[stream]               = false;
      
      param_i_stream_type[stream]     = 0;
      param_i_overlap_control[stream] = 0;
      param_i_map_type[stream]        = 0;
      param_i_param_type[stream]      = 0;
      param_i_param_control[stream]   = 0;
    }
    bprevframestatus = false;
    benableskip = false;

    char *char_buf = (char*)malloc(sizeof(int));
    if(char_buf == NULL){ throw new NodeException(NULL, string("char buffer can not allocated."), __FILE__, __LINE__); }

    header_reserved1 = 0;
    header_version   = 1;              // Version 1
    header_sz_char   = sizeof(char);   // Size of char
    header_sz_int    = sizeof(int);    // Size of int
    header_sz_float  = sizeof(float);  // Size of float   ** Size of complex float considers (float * 2)
    header_sz_double = sizeof(double); // Size of double  ** Size of complex double considers (double * 2)
    header_reserved2 = 0;

    // input_sourceID = addInput("IN_SOURCES"); // translateInput, because Matrix type aren't need source.
    input_dataID    = addInput("IN_PCM");
    output_sourceID = addOutput("OUT_SOURCES"); // if input was Matrix, always output no localize source.
    output_dataID   = addOutput("OUT_FLAC");

    string param_s_temp  = "";
    string param_s_temp0 = "";
    string param_s_temp1 = "";
    string param_s_temp2 = "";
    string param_s_temp3 = "";
    string param_s_temp4 = "";
    param_b_export = dereference_cast<bool>(parameters.get("ENABLE_EXPORT_DEBUG"));
    // Only enable when the ENABLE_EXPORT is ture.
    if(parameters.exist("EXPORT_FILENAME")){
      param_s_export_filename = object_cast<String>(parameters.get("EXPORT_FILENAME"));
      if(param_s_export_filename.c_str() == '\0'){ throw new NodeException(NULL, string("EXPORT_FILENAME is empty."), __FILE__, __LINE__); }
      if(param_b_export){
        // Open export file
        output_file_stream.open(param_s_export_filename.c_str());
        if(output_file_stream.fail()){ throw new NodeException(NULL, string("Can't open file'") + param_s_export_filename + "'.", __FILE__, __LINE__); }
        bstream_opend = true;
      }
    }
    else{
      param_s_export_filename = "";
    }
    param_b_debug_status = dereference_cast<bool>(parameters.get("ENABLE_STATUS_DEBUG"));
    param_s_temp = object_cast<String>(parameter.get("OUTPUT_UNIT"));
    if(     param_s_temp == "FRAME" ){ param_i_output_unit = 0; }
    else if(param_s_temp == "SOURCE"){ param_i_output_unit = 1; }
    else{                              param_i_output_unit = 1; }
    param_i_length = 0;  param_i_length = dereference_cast<int>(parameters.get("LENGTH"));
    if(param_i_length == 0){ param_i_length = 512; cerr << "Parameter error : LENGTH(no value or equal zero) change to 512." << endl; }
    param_i_advance = 0; param_i_advance = dereference_cast<int>(parameters.get("ADVANCE"));
    if(param_i_advance == 0){ param_i_advance = param_i_length; cerr << "Parameter error : ADVANCE(no value or equal zero) changed to as same as LENGTH value." << endl; }
    param_i_channels = dereference_cast<int>(parameters.get("CHANNEL"));
    param_i_sampling_rate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
    param_s_temp = object_cast<String>(parameter.get("SOURCE_HANDLING"));
    if(     param_s_temp == "FirstFrame"){  }
    else if(param_s_temp == "LastFrame" ){  }
    else if(param_s_temp == "MostStayed"){  }
    else if(param_s_temp == "Average"   ){  }
    else{                                   }
    param_s_temp = object_cast<String>(parameter.get("MATRIX_TO_MAP_TYPE"));
    if(     param_s_temp == "MultiChannel"){  }
    else if(param_s_temp == "EachChennel" ){  }
    else{                                     }

    param_b_mode_server = dereference_cast<bool>(parameters.get("SERVER_MODE"));
    param_b_mode_ssl = dereference_cast<bool>(parameters.get("USE_SSL"));
    param_b_mode_dosguard = dereference_cast<bool>(parameters.get("USE_DOS_ATTACK_GUARD"));
    // Only enable when the USE_DOS_ATTACK_GUARD is ture.
    if(parameters.exist("LIMIT")){ param_i_limit_size = dereference_cast<int>(parameters.get("LIMIT")); }
    else{                          param_i_limit_size = 0;                                              }
    param_s_host = object_cast<String>(parameters.get("PEER_HOST"));
    if(param_s_host.c_str() == '\0'){ throw new NodeException(NULL, string("PEER_HOST is empty."), __FILE__, __LINE__); }
    param_i_port = dereference_cast<int>(parameters.get("PEER_PORT"));
    param_s_dir = object_cast<String>(parameters.get("DIR_SUFFIX"));
    // Only enable when the SERVER_MODE is true(client).
    if(parameters.exist("USE_TIMEOUT")){ param_b_timeout = dereference_cast<bool>(parameters.get("USE_TIMEOUT")); }
    else{                                param_b_timeout = false;                                                 }
    // Only enable when the USE_TIMEOUT is ture.
    if(parameters.exist("TIMEOUT")){ param_i_timeout = dereference_cast<int>(parameters.get("TIMEOUT")); }
    else{                            param_i_timeout = 0;                                                }
    param_i_buffer = dereference_cast<int>(parameters.get("BUFFER_SIZE"));
    param_i_streams = dereference_cast<int>(parameters.get("NUM_OF_STREAM"));

    param_vi_skip_frame_trigger_streams = object_cast<Vector<int> >(parameters.get("INTELLIGENCE_FRAME_SKIP_TRIGER"));
    for(int stream = 0; stream < param_vi_skip_frame_trigger_streams.size(); stream++){
      if(!(param_vi_skip_frame_trigger_streams[stream] >= 0 && param_vi_skip_frame_trigger_streams[stream] < 16)){
        throw new NodeException(this, string("Noexistent stream is selectred. Please check the parameter INTELLIGENCE_FRAME"), __FILE__, __LINE__);
      }
      else{
        bskipcheckstream[param_vi_skip_frame_trigger_streams[stream]] = true;
        benableskip = true;
      }
    }

    for(int stream=0; stream<16; stream++){
      switch(stream){
      }
      if(parameters.exist(param_s_temp0)){
        param_s_temp = object_cast<String>(parameters.get(param_s_temp0));
        if(param_s_temp == "DataStream"){
          param_i_stream_type[stream] = 0; // input_type = 0x10000000;
          if(parameters.exist(param_s_temp1)){
            param_s_temp = object_cast<String>(parameters.get(param_s_temp1));
            if(param_s_temp == "Through"){         param_i_overlap_control[stream] = 0; } // default
            else if(param_s_temp == "AddFlag"){    param_i_overlap_control[stream] = 1; }
            else if(param_s_temp == "Overlap"){    param_i_overlap_control[stream] = 2; }
            else if(param_s_temp == "DisOverlap"){ param_i_overlap_control[stream] = 3; }
          }
          if(parameters.exist(param_s_temp2)){
            param_s_temp = object_cast<String>(parameters.get(param_s_temp2));
            if(param_s_temp == "MapIsFloat"){             param_i_map_type[stream] = 0; } // default
            else if(param_s_temp == "MapIsComplexFloat"){ param_i_map_type[stream] = 1; }
          }
        }
        else if(param_s_temp == "ParamStream"){
          param_i_stream_type[stream] = 1; // input_type = 0x20000000; // b_stream1_isdata = false;
          if(parameters.exist(param_s_temp3)){
            param_s_temp = object_cast<String>(parameters.get(param_s_temp3));
            if(param_s_temp == "Bool"){                     param_i_param_type[stream] = 0; } // default
            else if(param_s_temp == "Int"){                 param_i_param_type[stream] = 1; }
            else if(param_s_temp == "Float"){               param_i_param_type[stream] = 2; }
            else if(param_s_temp == "Double"){              param_i_param_type[stream] = 3; }
            else if(param_s_temp == "String"){              param_i_param_type[stream] = 4; }
            else if(param_s_temp == "Direct"){              param_i_param_type[stream] = 5; }
            else if(param_s_temp == "ParamsLocalizeMUSIC"){ param_i_param_type[stream] = 6; }
            else if(param_s_temp == "ParamsSourceTracker"){ param_i_param_type[stream] = 7; }
            else if(param_s_temp == "ParamsHRLE"){          param_i_param_type[stream] = 8; }
          }
          if(parameters.exist(param_s_temp4)){
            param_s_temp = object_cast<String>(parameters.get(param_s_temp4));
            if(param_s_temp == "HoldAndUpdate"){              param_i_param_control[stream] = 0; } // default
            else if(param_s_temp == "HoldAndSmoothlyUpdate"){ param_i_param_control[stream] = 1; }
            else if(param_s_temp == "LedInitial"){            param_i_param_control[stream] = 2; }
            else if(param_s_temp == "ZeroInit"){              param_i_param_control[stream] = 3; }
            else if(param_s_temp == "OneInit"){               param_i_param_control[stream] = 4; }
            else if(param_s_temp == "LikeGeyser"){            param_i_param_control[stream] = 5; }
          }
        }
      }
    }

    if(param_i_streams <   1){ cerr << "Parameter error : NUM_OF_STREAM(Out of range) " << param_i_streams << endl; param_i_streams = 1; }
    if(param_i_streams >  16){ cerr << "Parameter error : NUM_OF_STREAM(Out of range) " << param_i_streams << endl; param_i_streams = 16; }
    for(int stream=param_i_streams-1; stream>=0; stream--){ binputProc[stream]  = true; }

    free(char_buf);

    // Start connection (WebSocket open)
    if(param_b_mode_server){
      try{
        //// create an instance of handler ////
        //websocketconnector::server_handler::startServer(); // temp function define. normally ****->startServer();
        //server::handler::ptr handler(new ws_server_handler()); 
        //ws_server_handler_ptr handler(new ws_server_handler());
        m_srv_handler.reset(new websocketconnector::ws_server_handler());

        //// create a server the listen on port uses handler ////
        //websocketpp::server endpoint(handler);
        m_srv_endpoint = (new websocketpp::server(m_srv_handler));

        // set log level for message print
        //srv_endpoint.alog().set_level(websocketpp::log::alevel::CONNECT);
        //srv_endpoint.alog().set_level(websocketpp::log::alevel::DISCONNECT);
        //srv_endpoint.elog().set_level(websocketpp::log::elevel::RERROR);
        //srv_endpoint.elog().set_level(websocketpp::log::elevel::FATAL);
        m_srv_endpoint->alog().set_level(websocketpp::log::alevel::CONNECT);
        m_srv_endpoint->alog().set_level(websocketpp::log::alevel::DISCONNECT);
        m_srv_endpoint->elog().set_level(websocketpp::log::elevel::RERROR);
        m_srv_endpoint->elog().set_level(websocketpp::log::elevel::FATAL);

        if(param_i_limit_size > 0){
          //server->set_max_messages_size(param_i_limit_size);
          std::cout << "Set receive message limitation [" << param_i_limit_size << " Byte]." << std::endl;
        }

        //m_srv_endpoint->listen(param_i_port);
        //std::cout << "Starting WebSocket server on port [" << param_i_port << "]." << std::endl;

        //// start thread ////
        //boost::thread thread_hark(boost::bind(&client::run, &cli_endpoint, false));
        //m_thread_hark.reset(new boost::thread(boost::bind(&websocketpp::role::server<websocketpp::server>::listen, m_srv_endpoint, param_i_port, 1)));
        void(websocketpp::role::server<websocketpp::server>::*f)(uint16_t, size_t) = &websocketpp::role::server<websocketpp::server>::listen;
        m_thread_hark.reset(new boost::thread(boost::bind(f, m_srv_endpoint, param_i_port, 1)));
        std::cout << "Starting WebSocket server on port [" << param_i_port << "]." << std::endl;
      }
      catch(exception& e){
        std::cerr << "Exception: " << e.what() << std::endl;
      }
    }
    else{
      std::stringstream ss_uri;
      ss_uri << "ws://" << param_s_host << ":" << param_i_port << "/" << param_s_dir;
      std::string uri = ss_uri.str();
      //std::string uri = "ws://localhost:8003/hark";
      try{
        //// create an instance of handler ////
        //websocketconnector::client_handler::startClient(); // temp function define. normally ****->startClient();
        //websocketconnector::ws_client_handler_ptr cli_handler(new websocketconnector::ws_client_handler());
        //client::handler::ptr handler(new websocketconnector::ws_client_handler());
        m_cli_handler.reset(new websocketconnector::ws_client_handler());

        //// create a client uses handler ////
        //websocketpp::client cli_endpoint(cli_handler);
        m_cli_endpoint = (new websocketpp::client(m_cli_handler));

        //websocketpp::client::connection_ptr connect_ptr;

        //// set log level for message print ////
        //cli_endpoint.alog().unset_level(websocketpp::log::alevel::ALL); // disable all infomation message
        //cli_endpoint.elog().unset_level(websocketpp::log::elevel::ALL); // disable all error message
        //cli_endpoint.elog().set_level(websocketpp::log::elevel::RERROR);
        //cli_endpoint.elog().set_level(websocketpp::log::elevel::FATAL);
        m_cli_endpoint->alog().unset_level(websocketpp::log::alevel::ALL); // disable all infomation message
        m_cli_endpoint->elog().unset_level(websocketpp::log::elevel::ALL); // disable all error message
        m_cli_endpoint->elog().set_level(websocketpp::log::elevel::RERROR);
        m_cli_endpoint->elog().set_level(websocketpp::log::elevel::FATAL);

        //// get connection ////
        //connect_ptr = cli_endpoint.get_connection(uri);
        //connect_ptr->add_request_header("User-Agent", "WebSocket++/0.2.0 FlowDesignerPublisherWebSocket/0.0.1");
        //connect_ptr->add_subprotocol("hri.websocketpp.hark");
        //connect_ptr->set_origin("http://hark.jp.honda-ri.com");
        //cli_endpoint.connect(connect_ptr);
        m_connect_ptr = m_cli_endpoint->get_connection(uri);
        m_connect_ptr->add_request_header("User-Agent", "WebSocket++/0.2.0 FlowDesignerPublisherWebSocket/0.0.1");
        m_connect_ptr->set_origin("http://hark.jp.honda-ri.com");
        m_cli_endpoint->connect(m_connect_ptr);

        //// start thread ////
        //boost::thread thread_hark(boost::bind(&client::run, &cli_endpoint, false));
        m_thread_hark.reset(new boost::thread(boost::bind(&client::run, m_cli_endpoint, false)));
        std::cout << "Starting WebSocket client." << std::endl;

        //// send test code (obsolete) ////
        //char test_buffer[2048];
        //while(cin.getline(test_buffer, 2048)){ cli_handler->send(test_buffer); }
        //// waiting thread finish for send test code (obsolete) ////
        //thread_hark.join();
      }
      catch(std::exception& e){
        std::cerr << "Exception: " << e.what() << std::endl;
      }
    }

    //buf_size = 16000 * 16 * 10; buffer = (short int*)calloc(sizeof(short int), buf_size);
    //c_sock = socket(AF_INET, SOCK_STREAM, 0);
    //if(c_sock < 0){ perror("building socket failed"); exit(-1); }
    //addr.sin_family = AF_INET;
    //addr.sin_port = htons(port);
    //addr.sin_addr.s_addr = INADDR_ANY;
    //setsockopt(c_sock, SOL_SOCKET, SO_REUSEADDR, (const char*)&yes, sizeof(yes)); yes = 1;
    //test2_client();
    inOrder = true;
    std::cout << getName() << " Constructed..." << std::endl;
  }

  ~PublisherWebSocket()
  {
    // Close export file
    if(bstream_opend){ output_file_stream.close(); }

    // Stop connection (WebSocket close)
    if(param_b_mode_server){
      //websocketconnector::ws_server_handler::stopServer(); // temp function define. normally ****->stopServer();
      if(m_thread_hark){ m_thread_hark->join(); }
    }
    else{
      //websocketconnector::ws_client_handler::stopClient(); // temp function define. normally ****->stopClient();
      if(m_thread_hark){ m_thread_hark->join(); }
    }

    
    
    cout << getName() << " Destructed..." << endl;
  }


  // dynamic input-port translation
  virtual int translateInput(string inputName) {
    // check whether the input ports have arc from other module.
    for(unsigned int i = 0; i < inputs.size(); i++) {
      if(inputs[i].name == inputName) {
        return i;
      }
    }
    if(inputName == "IN_0"){       binputID[0]  = true; return inputID[0]  = addInput(inputName); }
    else if(inputName == "IN_1"){  binputID[1]  = true; return inputID[1]  = addInput(inputName); }
    else if(inputName == "IN_2"){  binputID[2]  = true; return inputID[2]  = addInput(inputName); }
    else if(inputName == "IN_3"){  binputID[3]  = true; return inputID[3]  = addInput(inputName); }
    else if(inputName == "IN_4"){  binputID[4]  = true; return inputID[4]  = addInput(inputName); }
    else if(inputName == "IN_5"){  binputID[5]  = true; return inputID[5]  = addInput(inputName); }
    else if(inputName == "IN_6"){  binputID[6]  = true; return inputID[6]  = addInput(inputName); }
    else if(inputName == "IN_7"){  binputID[7]  = true; return inputID[7]  = addInput(inputName); }
    else if(inputName == "IN_8"){  binputID[8]  = true; return inputID[8]  = addInput(inputName); }
    else if(inputName == "IN_9"){  binputID[9]  = true; return inputID[9]  = addInput(inputName); }
    else if(inputName == "IN_10"){ binputID[10] = true; return inputID[10] = addInput(inputName); }
    else if(inputName == "IN_11"){ binputID[11] = true; return inputID[11] = addInput(inputName); }
    else if(inputName == "IN_12"){ binputID[12] = true; return inputID[12] = addInput(inputName); }
    else if(inputName == "IN_13"){ binputID[13] = true; return inputID[13] = addInput(inputName); }
    else if(inputName == "IN_14"){ binputID[14] = true; return inputID[14] = addInput(inputName); }
    else if(inputName == "IN_15"){ binputID[15] = true; return inputID[15] = addInput(inputName); }
    else if(inputName == "IN_TRANSMISSION_CONTROL"){ binput_transmission_controlID = true; return input_transmission_controlID = addInput(inputName); }
    else if(inputName == "IN_CONDITION_SYNC"){ binput_condition_syncID = true; return input_condition_syncID = addInput(inputName); }
    else{ throw new NodeException(this, inputName + " is not supported.", __FILE__, __LINE__); }
  }

  // Initialize
  virtual void initialize() {

    outputs[output_conditionID].lookAhead = 
    outputs[output_dummyID].lookAhead = 
    
    1+max(outputs[output_conditionID].lookAhead,
          outputs[output_dummyID].lookAhead);

    outputs[output_conditionID].lookAhead = 
    outputs[output_dummyID].lookAhead = 
    
    1+max(outputs[output_conditionID].lookBack,
          outputs[output_dummyID].lookBack);

    // I know that it is not desirable that follows "this->BufferedNode::initialize();".
    // However, "this->BufferedNode::initialize();" performed as a work-around for slow processing.
    int countout = 0;
    if(param_b_mode_server){
      // Server listen loop recommend infinite loop. If you need terminative loop by timeout, change to use this lines.
      //std::cout << "waiting client connection...";
      //while(!(m_srv_handler->get_connect_state())){ std::cout << "."; countout++; if(countout > 0x00FFFFFF){ throw new NodeException(this, "connection timeout.", __FILE__, __LINE__); } }
      //std::cout << std::endl;
      std::cout << "waiting client connection...";
      while(!(m_srv_handler->get_connect_state())){ std::cout << "."; }
      std::cout << std::endl;
    }
    else{
      std::cout << "waiting server connection...";
      while(!(m_cli_handler->get_connect_state())){ std::cout << "."; countout++; if(countout > 0x00FFFFFF){ throw new NodeException(this, "connection timeout.", __FILE__, __LINE__); } }
      std::cout << std::endl;
    }

    this->BufferedNode::initialize();
    std::cout << getName() << " Initialized..." << std::endl;
  }

  inline int32_t checkClipAndFloatToInt32(float x, int bytes_per_sample)
  {
    int32_t ret=0;
    switch(bytes_per_sample){
    case 2:
      ret = (int32_t)(x >= SHRT_MAX ? SHRT_MAX : x);
      ret = (int32_t)(x <= SHRT_MIN ? SHRT_MIN : x);
      break;
    case 3:
      ret = (int32_t)(x >= (((SHRT_MAX + 1) * (UCHAR_MAX + 1)) - 1) ? (((SHRT_MAX + 1) * (UCHAR_MAX + 1)) - 1) : x);
      ret = (int32_t)(x <= (-(SHRT_MAX + 1) * (UCHAR_MAX + 1))      ? (-(SHRT_MAX + 1) * (UCHAR_MAX + 1))      : x);
      break;
    case 4:
      ret = (int32_t)(x > INT_MAX - 100 ? INT_MAX - 100 : x); // fix me
      ret = (int32_t)(x < INT_MIN ? INT_MIN : x);
      break;
    }
    return ret;
  }


  // memo:FLACの処理にint32_tは必須なのでint型とint32_tの切り替えは不要
  void toCharVector(Matrix<float>& x, vector<char>& buf, int t_first, int t_last)
  {
    buf.resize(x.nrows() * (t_last - t_first) * bytes_per_sample);
    vector<char>::iterator it = buf.begin();
    
    for(int j=t_first; j<t_last; j++){
      for(int i = 0; i < x.nrows(); i++){
        int32_t v = checkClipAndFloatToInt32(x(i, j),bytes_per_sample);
        char* p = (char*)&v;
        copy(p, p + bytes_per_sample, it);
        it += bytes_per_sample;
      }
    }
    if(it != buf.end()){
      cout << distance(it, buf.end()) << endl;
    }
  }

  void toCharVector(Vector<float>& x, vector<char>& buf, int t_first, int t_last)
  {
    buf.resize((t_last - t_first) * bytes_per_sample);
    vector<char>::iterator it = buf.begin();
    
    for(int i=t_first; i<t_last; i++){
      int32_t v = checkClipAndFloatToInt32(x[i],bytes_per_sample);
      char* p = (char*)&v;
      copy(p, p + bytes_per_sample, it);
      it += bytes_per_sample;
    }
  }

  // Implementation
  void calculate(int output_id, int count, Buffer &out){
    
//    // update websocket condition flag
//    if(param_b_mode_server){
//      if(m_thread_hark){
//        if(m_srv_handler->get_connect_state()){ socket_condition = true;  }
//        else{                                   socket_condition = false; }
//      }
//      else{
//        socket_condition = false;
//      }
//    }
//    else{
//      if(m_thread_hark){
//        if(m_cli_handler->get_connect_state()){ socket_condition = true;  }
//        else{                                   socket_condition = false; }
//      }
//      else{
//        socket_condition = false;
//      }
//    }
//    // control condition flag
//    Buffer &output_conditionBuffer = *(outputs[output_conditionID].buffer);
//    Buffer &output_dummyBuffer = *(outputs[output_dummyID].buffer);
//    if(binput_condition_syncID){
//      ObjectRef input_condition_obj = getInput(input_condition_syncID, count);
//      bool input_condition = dereference_cast<bool> (input_condition_obj);
//      if(input_condition && socket_condition){ output_conditionBuffer[count] = TrueObject;  }
//      else{                                    output_conditionBuffer[count] = FalseObject; }
//      if(input_condition){ output_dummyBuffer[count] = TrueObject;  }
//      else{                output_dummyBuffer[count] = FalseObject; }
//    }
//    else{
//      output_dummyBuffer[count] = FalseObject;
//      if(socket_condition){ output_conditionBuffer[count] = TrueObject;  }
//      else{                 output_conditionBuffer[count] = FalseObject; }
//    }
//
//
//    bool transmission_completion_flag = false;
//    bool no_need_transmission = false;
//
//    string buffer("");
//    buffer.append(header_string.begin(), header_string.end());
//    for(int i=0; i<header_sz_int; i++){ buffer.push_back((reinterpret_cast<char*>(&param_i_streams))[i]); }


    
    struct flac_encoder {
      Source src;
      
    };
    //std::map<int, flac_encoder> encoder_map;
    
    RCPtr<Vector<ObjectRef> > sources_ptr = getInput(sourcesID, count);
    const Vector<ObjectRef>& sources = *sources_ptr;
    
    RCPtr<Map<int, ObjectRef> > output(new Map<int, ObjectRef>);
    
    for(int i=0; i<sources.size(); i++){
      RCPtr<Source> src = sources[i];

      Map<int, ObjectRef>::const_iterator feature_it = features.find(src->id);
      Map<int, ObjectRef>::const_iterator mask_it = masks.find(src->id);
      assert(feature_it != features.end() && mask_it != masks.end());

      if(encoder_map.find(src->id) == encoder_map.end()) { // New Source ID
        flac_encoder& new_encoder = encoder_map[src->id];
        new_encoder.src_info.source_id = src->id;
        new_param.src_info.azimuth = 180 / M_PI * atan2f(src->x[1], src->x[0]);
        new_param.src_info.elevation = 180 / M_PI * asin(src->x[2]);
        timeval tv;
        gettimeofday(&tv, NULL);
        new_param.src_info.tv_sec = tv.tv_sec;
        new_param.src_info.tv_usec = tv.tv_usec;

        new_param.feature_list = new Vector<FVectorPtr>;
        new_param.mask_list = new Vector<FVectorPtr>;
        new_param.spec_mean = new Vector<float>(object_cast<Vector<float> >(feature_it->second).size() / 2);
      }
      asr_param& param = param_map[src->id];
            
      param.feature_list->push_back(feature_it->second);
      param.mask_list->push_back(mask_it->second);
      Vector<float>& feature = object_cast<Vector<float> >(feature_it->second);
      for (int i = 0; i < feature.size() / 2; i++) {
        (*param.spec_mean)[i] += object_cast<Vector<float> >(feature_it->second)[i];
      }
    }

    map<int, flac_encoder>::iterator it = encoder_map.begin();
    while(it != encoder_map.end()){
      if(features.find(it->first) == features.end()){
        sendASRParam(it->second);
        encoder_map.erase(it++);
      }
      else {
        ++it;
      }
    }
        
    out[count] = sources_ptr;
    
    
    



    Buffer &output_sourceBuffer = *(outputs[output_sourceID].buffer);
    Buffer &output_dataBuffer = *(outputs[output_dataID].buffer);
    
    if(binput_sourceID){ // use source :: The source ID handling like SpeachRecognitionSMNClient. Only processing target source ID.
      ObjectRef in_source = getInput(input_sourceID, count);
      ObjectRef in_data = getInput(input_dataID, count);
      
      
    }
    else{ // not use source :: The source ID handling like SaveRawPCM. Processing the source ID of all.
      ObjectRef in_data = getInput(input_dataID, count);
      
      if(typeid(*in_data) == typeid(Map<int, ObjectRef>)){
        RCPtr<Map<int, ObjectRef> > output_map_p(new Map<int, ObjectRef>); // 出力バッファ
        output_dataBuffer[count] = output_map_p;
        Map<int, ObjectRef>& output = *output_map_p;
        
        RCPtr<Map<int, ObjectRef> > input = in_data;
        Map<int, ObjectRef>::iterator it = input->begin();
        for(; it != input->end(); ++it){
          Vector<float>& x = object_cast<Vector<float> >(it->second);
          
          // Confirm matching in the already existing source ID.
          if(streams.find(it->first) == streams.end()){ // In the case of the source ID of first frame.
            if(file_count < it->first){ file_count = it->first; }
            ostringstream name
            toCharVector(x, buf, 0, x.size());
          }
          else{ // In the case of the source ID of second or later frame.
            toCharVector(x, buf, x.size() - advance, x.size());
          }
        }
        
        map<int, flac_encoder>::iterator it = encoder_map.begin();
        while(it != encoder_map.end()){
          if(input->find(it->first) == input->end()){ // Terminated source ID was found.
            RCPtr<Vector<int> > vec_int(new Vector<int>(vec_size)); // vec_size is length
            output[map_key] = vec_int; // map_key is sorce ID.
            ENCODER->send(it->second);
            encoder_map.erase(it++);
          }
          else {
            ++it;
          }
        }
      }
      else if(typeid(*in_data) == typeid(Matrix<float>)){
        RCPtr<Vector<int> > output_vector_p(new Vector<int>(vec_size)); // 出力バッファ
        output_dataBuffer[count] = output_vector_p;
        Vector<int>& output = *output_vector_p;
      
        RCPtr<Matrix<float> > output_matrix_p(new Matrix<float>(rows, cols)); // 出力バッファ
        output_dataBuffer[count] = output_matrix_p;
        Matrix<float>& output = *output_matrix_p;
        
        Matrix<float>& x = object_cast<Matrix<float> >(in_data);
        int mat_num_ch = x.nrows();
        int mat_length = x.ncols();
        if(!fout){
          fout.reset(new ofstream((basename + ".sw").c_str(), ios::binary));
          toCharVector(x, buf, 0, mat_length - advance);
          ENCODER->write((char*)&buf[0], buf.size());
          ENCODER->flush();
        }
        toCharVector(x, buf, mat_length - advance, mat_length);
        ENCODER->write((char*)&buf[0], buf.size());
        ENCODER->flush();
      }
      else{
        throw new NodeException(this, string("Only Map<int, ObjectRef> and Matrix<float> is supported."), __FILE__, __LINE__);
      }
      out[count] = in_data;
    }




































    for(int stream=0; stream<16; stream++){
      if(binputProc[stream]){
        string stream_buffer("");
        int n_stream = stream;//+1; //0x00000001; // Stream No.   [ 0 ... 15 ]
        if(binputID[stream]){
          ObjectRef input_ptr; input_ptr = getInput(inputID[stream], count);
          // Buffer &outputBuffer[stream]  = *(outputs[outputID[stream]].buffer);  outputBuffer[stream][count]  = input_ptr[stream];
          if(typeid(*input_ptr) == typeid(Vector<ObjectRef>)){
            int s_header = 0x01020F00; // Start Header [ 0x01:Start 0x020F:Vector<ObjectRef> 0x00:End ]
            int e_header = 0x11020F10; // End Header   [ 0x11:Start 0x020F:Vector<ObjectRef> 0x10:End ]
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
            RCPtr<Vector<ObjectRef> > src_type_p = getInput(inputID[stream], count);
            const Vector<ObjectRef>& sources = *src_type_p;
            int src_size = sources.size();
            //cout << "Sources: " << src_size << " Data: ";
            bisnodata[stream] = (src_size == 0 ? true : false);
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&src_size))[i]); }
            for(int i=0; i<src_size; i++){
              RCPtr<Source> src = sources[i];
              float float_tmp;
              int int_tmp;
              int_tmp = src->id;
              for(int j=0; j<header_sz_int; j++){
                stream_buffer.push_back((reinterpret_cast<char*>(&int_tmp))[j]);
              }
              float_tmp = src->x[0];
              for(int j=0; j<header_sz_float; j++){
                stream_buffer.push_back((reinterpret_cast<char*>(&float_tmp))[j]);
              }
              float_tmp = src->x[1];
              for(int j=0; j<header_sz_float; j++){
                stream_buffer.push_back((reinterpret_cast<char*>(&float_tmp))[j]);
              }
              float_tmp = src->x[2];
              for(int j=0; j<header_sz_float; j++){
                stream_buffer.push_back((reinterpret_cast<char*>(&float_tmp))[j]);
              }
              float_tmp = src->power;
              for(int j=0; j<header_sz_int; j++){
                stream_buffer.push_back((reinterpret_cast<char*>(&float_tmp))[j]);
              }
              //cout << count << " " << src->id << " " << src->x[0] << " " << src->x[1] << " " << src->x[2] << " " << src->power << " " << endl;
              //cout << stream_buffer << endl;
            }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
            cout << "Stream" << n_stream << " : Vector<ObjectRef>(Source) : " << endl;
          }

          else if(typeid(*input_ptr) == typeid(Vector<int>)){
            int s_header = 0x01020300; // Start Header [ 0x01:Start 0x0203:Vector<int> 0x00:End ]
            int e_header = 0x11020310; // End Header   [ 0x11:Start 0x0203:Vector<int> 0x10:End ]
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
            RCPtr<Vector<int> > input = getInput(inputID[stream], count);
            int vec_size = (*input).size();
            //cout << "VecSize: " << vec_size << " Data: ";
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&vec_size))[i]); }
            for(int i=0; i<vec_size; i++){
              for(int j=0; j<header_sz_int; j++){
                stream_buffer.push_back((reinterpret_cast<char*>(&((*input)[i])))[j]);
              }
            }
            //cout << stream_buffer << endl;
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
            cout << "Stream" << n_stream << " : Vector<int> : " << endl;
          }

          else if(typeid(*input_ptr) == typeid(Vector<float>)){
            int s_header = 0x01020500; // Start Header [ 0x01:Start 0x0205:Vector<float> 0x00:End ]
            int e_header = 0x11020510; // End Header   [ 0x11:Start 0x0205:Vector<float> 0x10:End ]
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
            RCPtr<Vector<float> > input = getInput(inputID[stream], count);
            int vec_size = (*input).size();
            //cout << "VecSize: " << vec_size << " Data: ";
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&vec_size))[i]); }
            for(int i=0; i<vec_size; i++){
              for(int j=0; j<header_sz_float; j++){
                stream_buffer.push_back((reinterpret_cast<char*>(&((*input)[i])))[j]);
              }
            }
            //cout << stream_buffer << endl;
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
            cout << "Stream" << n_stream << " : Vector<float> : " << endl;
          }

          else if(typeid(*input_ptr) == typeid(Vector<complex<float> >)){
            int s_header = 0x01020600; // Start Header [ 0x01:Start 0x0206:Vector<complex<float> > 0x00:End ]
            int e_header = 0x11020610; // End Header   [ 0x11:Start 0x0206:Vector<complex<float> > 0x10:End ]
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
            RCPtr<Vector<complex<float> > > input = getInput(inputID[stream], count);
            int vec_size = (*input).size();
            //cout << "VecSize: " << vec_size << " Data: ";
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&vec_size))[i]); }
            for(int i=0; i<vec_size; i++){
              for(int j=0; j<header_sz_float; j++){
                stream_buffer.push_back((reinterpret_cast<char*>(&((*input)[i].real())))[j]);
              }
              for(int j=0; j<header_sz_float; j++){
                stream_buffer.push_back((reinterpret_cast<char*>(&((*input)[i].imag())))[j]);
              }
            }
            //cout << stream_buffer << endl;
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
            cout << "Stream" << n_stream << " : Vector<complex<float> > : " << endl;
          }

          else if(typeid(*input_ptr) == typeid(Matrix<int>)){
            int s_header = 0x01030300; // Start Header [ 0x01:Start 0x0303:Matrix<int> 0x00:End ]
            int e_header = 0x11030310; // End Header   [ 0x11:Start 0x0303:Matrix<int> 0x10:End ]
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
            RCPtr<Matrix<int> > input = getInput(inputID[stream], count);
            int mat_size_row = (*input).nrows();
            int mat_size_col = (*input).ncols();
            //cout << "MatrixRows: " << mat_size_row << " MatrixColum: " << mat_size_col << " Data: ";
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&mat_size_row))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&mat_size_col))[i]); }
            for(int row=0; row<mat_size_row; row++){
              for(int col=0; col<mat_size_col; col++){
                for(int j=0; j<header_sz_int; j++){
                  stream_buffer.push_back((reinterpret_cast<char*>(&((*input)(row, col))))[j]);
                }
              }
            }
            //cout << stream_buffer << endl;
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
            cout << "Stream" << n_stream << " : Matrix<int> : " << endl;
          }

          else if(typeid(*input_ptr) == typeid(Matrix<float>)){
            int s_header = 0x01030500; // Start Header [ 0x01:Start 0x0305:Matrix<float> 0x00:End ]
            int e_header = 0x11030510; // End Header   [ 0x11:Start 0x0305:Matrix<float> 0x10:End ]
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
            RCPtr<Matrix<float> > input = getInput(inputID[stream], count);
            int mat_size_row = (*input).nrows();
            int mat_size_col = (*input).ncols();
            //cout << "MatrixRows: " << mat_size_row << " MatrixColum: " << mat_size_col << " Data: ";
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&mat_size_row))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&mat_size_col))[i]); }
            for(int row=0; row<mat_size_row; row++){
              for(int col=0; col<mat_size_col; col++){
                for(int j=0; j<header_sz_float; j++){
                  stream_buffer.push_back((reinterpret_cast<char*>(&((*input)(row, col))))[j]);
                }
              }
            }
            //cout << stream_buffer << endl;
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
            cout << "Stream" << n_stream << " : Matrix<float> : " << endl;
          }

          else if(typeid(*input_ptr) == typeid(Matrix<complex<float> >)){
            int s_header = 0x01030600; // Start Header [ 0x01:Start 0x0306:Matrix<complex<float> > 0x00:End ]
            int e_header = 0x11030610; // End Header   [ 0x11:Start 0x0306:Matrix<complex<float> > 0x10:End ]
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
            RCPtr<Matrix<complex<float> > > input = getInput(inputID[stream], count);
            int mat_size_row = (*input).nrows();
            int mat_size_col = (*input).ncols();
            //cout << "MatrixRows: " << mat_size_row << " MatrixColum: " << mat_size_col << " Data: ";
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&mat_size_row))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&mat_size_col))[i]); }
            for(int row=0; row<(*input).nrows(); row++){
              for(int col=0; col<(*input).ncols(); col++){
                for(int j=0; j<header_sz_float; j++){
                  stream_buffer.push_back((reinterpret_cast<char*>(&((*input)(row, col).real())))[j]);
                }
                for(int j=0; j<header_sz_float; j++){
                  stream_buffer.push_back((reinterpret_cast<char*>(&((*input)(row, col).imag())))[j]);
                }
              }
            }
            //cout << stream_buffer << endl;
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
            cout << "Stream" << n_stream << " : Matrix<complex<float> > : " << endl;
          }

          else if(typeid(*input_ptr) == typeid(Map<int, ObjectRef>)){
            int s_header = 0x01040F00; // Start Header [ 0x01:Start 0x040F:Map<int, ObjectRef> 0x00:End ]
            int e_header = 0x11040F10; // End Header   [ 0x11:Start 0x040F:Map<int, ObjectRef> 0x10:End ]
            if(param_i_map_type[stream] == 0){
              s_header = 0x01040500; // Overwrite Start Header [ 0x01:Start 0x0405:Map<int, Vector<float> > 0x00:End ]
              e_header = 0x11040510; // Overwrite End Header   [ 0x11:Start 0x0405:Map<int, Vector<float> > 0x10:End ]
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
              RCPtr<Map<int, ObjectRef> > input = getInput(inputID[stream], count);
              Map<int, ObjectRef>::const_iterator it;
              int map_size = (*input).size();
              bisnodata[stream] = (map_size == 0 ? true : false);
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&map_size))[i]); }
              switch(param_i_overlap_control[stream]){
              case 3: // DisOverlap mode
                for(it=input->begin(); it != input->end(); ++it){
                  Vector<float>& tmpvec = object_cast<Vector<float> >(it->second);
                  int map_key = it->first;
                  int vec_size = (tmpvec.size() < param_i_advance) ? tmpvec.size() : param_i_advance;
                  //cout << "MapKey: " << map_key << " MapVecSize: " << vec_size << " Data: ";
                  for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&map_key))[i]); }
                  for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&vec_size))[i]); }
                  for(int i=0; i<vec_size; i++){
                    for(int j=0; j<header_sz_float; j++){
                      stream_buffer.push_back((reinterpret_cast<char*>(&(tmpvec[i])))[j]);
                    }
                  }
                  //cout << stream_buffer << endl;
                }
                break;
              case 2: // Overlap mode
                break;
              case 1: // AddFlag mode
                break;
              case 0: // Through mode
              default:
                for(it=input->begin(); it != input->end(); ++it){
                  Vector<float>& tmpvec = object_cast<Vector<float> >(it->second);
                  int map_key = it->first;
                  int vec_size = tmpvec.size();
                  //cout << "MapKey: " << map_key << " MapVecSize: " << vec_size << " Data: ";
                  for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&map_key))[i]); }
                  for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&vec_size))[i]); }
                  for(int i=0; i<vec_size; i++){
                    for(int j=0; j<header_sz_float; j++){
                      stream_buffer.push_back((reinterpret_cast<char*>(&(tmpvec[i])))[j]);
                    }
                  }
                  //cout << stream_buffer << endl;
                }
                break;
              }
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
              cout << "Stream" << n_stream << " : Map<int, Vector<float> > : " << endl;
            }
            else if(param_i_map_type[stream] == 1){
              s_header = 0x01040600; // Overwrite Start Header [ 0x01:Start 0x0406:Map<int, Vector<complex<float> > > 0x00:End ]
              e_header = 0x11040610; // Overwrite End Header   [ 0x11:Start 0x0406:Map<int, Vector<complex<float> > > 0x10:End ]
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
              RCPtr<Map<int, ObjectRef> > input = getInput(inputID[stream], count);
              Map<int, ObjectRef>::const_iterator it;
              int map_size = (*input).size();
              bisnodata[stream] = (map_size == 0 ? true : false);
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&map_size))[i]); }
              for(it=input->begin(); it != input->end(); ++it){
                Vector<complex<float> >& tmpvec = object_cast<Vector<complex<float> > >(it->second);
                int map_key = it->first;
                int vec_size = tmpvec.size();
                for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&map_key))[i]); }
                for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&vec_size))[i]); }
                //cout << "MapKey: " << map_key << " MapVecSize: " << vec_size << " Data: ";
                for(int i=0; i<vec_size; i++){
                  for(int j=0; j<header_sz_float; j++){
                    stream_buffer.push_back((reinterpret_cast<char*>(&(tmpvec[i].real())))[j]);
                  }
                  for(int j=0; j<header_sz_float; j++){
                    stream_buffer.push_back((reinterpret_cast<char*>(&(tmpvec[i].imag())))[j]);
                  }
                }
                //cout << stream_buffer << endl;
              }
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
              cout << "Stream" << n_stream << " : Map<int, Vector<complex<float> > > : " << endl;
            }
            else{ // Map<int, ObjectRef> is unsupported now => zero size map.
              s_header = 0x01040F00; // Start Header [ 0x01:Start 0x040F:Map<int, ObjectRef> 0x00:End ]
              e_header = 0x11040F10; // End Header   [ 0x11:Start 0x040F:Map<int, ObjectRef> 0x10:End ]
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
              int map_size = 0;
              bisnodata[stream] = (map_size == 0 ? true : false);
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&map_size))[i]); }
              for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
              cout << "Stream" << n_stream << " : Map<int, ObjectRef> : " << endl;
            }
          }
          else{ // Unknown type is No data
            int s_header = 0x01000000; // Start Header [ 0x01:Start 0x0000:No data 0x00:End ]
            int e_header = 0x11000010; // End Header   [ 0x11:Start 0x0000:No data 0x10:End ]
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
            for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
            cout << "Stream" << n_stream << " : No data : " << endl;
          }
        }
        else{ // No input is No data
          int s_header = 0x01000000; // Start Header [ 0x01:Start 0x0000:No data 0x00:End ]
          int e_header = 0x11000010; // End Header   [ 0x11:Start 0x0000:No data 0x10:End ]
          for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&n_stream))[i]); }
          for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&s_header))[i]); }
          for(int i=0; i<header_sz_int; i++){ stream_buffer.push_back((reinterpret_cast<char*>(&e_header))[i]); }
          cout << "Stream" << n_stream << " : No data : " << endl;
        }
        buffer.append(stream_buffer.begin(), stream_buffer.end());
        hex_string_print(stream_buffer);
      }
    }

    if(binput_transmission_controlID){
      ObjectRef input_transmission_control_obj = getInput(input_transmission_controlID, count);
      bool input_transmission_flag = dereference_cast<bool> (input_transmission_control_obj);
      if(!input_transmission_flag){ no_need_transmission = true; }
    }

    if(!no_need_transmission){
      if(benableskip){ // If enable Intelligence frame skip function.
        bool btmpskip = true;
        int stream = 0;
        while(btmpskip && stream<param_i_streams){
          if(bskipcheckstream[stream] && !bisnodata[stream]){ btmpskip = false; } // If there is the data in the stream of check target, then not skip frame.
          stream++;
        }
        if(btmpskip && bprevframestatus){ no_need_transmission = true; } // If the skip frame flag detection is continued, then updates the transmission control flag.
        bprevframestatus = btmpskip; // backup skip frame flag for next frame process.
      }
    }

    
    if(!no_need_transmission){
      if(param_b_mode_server){
        //websocketconnector::ws_server_handler::send(buffer); // temp function define. normally ****->send(buffer);
        while(!transmission_completion_flag){ transmission_completion_flag = m_srv_handler->send(buffer); usleep(500); }
      }
      else{
        //websocketconnector::ws_client_handler::send(buffer); // temp function define. normally ****->send(buffer);
        while(!transmission_completion_flag){ transmission_completion_flag = m_cli_handler->send(buffer); usleep(500); }
        //[obsolete debug code] // websocketconnector::ws_client ssss; ssss.client_start("am3", 9003, "hark"); ssss.m_cli_handler->send(buffer); // (*m_cli_handler)->send(buffer); (*m_thread_hark).join();
      }
    }

    if(binput_condition_syncID){
      ObjectRef input_condition_obj = getInput(input_condition_syncID, count);
      bool input_condition = dereference_cast<bool> (input_condition_obj);
      if(input_condition && socket_condition){  }
      else{
        if(param_b_mode_server){ /*m_srv_endpoint->close();*/ }
        else{                    m_cli_handler->close(); }
      }
    }
    else{
      if(socket_condition){  }
      else{
        if(param_b_mode_server){ /*m_srv_endpoint->close();*/ }
        else{                    m_cli_handler->close(); }
      }
    }


/*
    if(enable_count){ fout << count << " "; }
    if(timestampID != -1){
      ObjectRef t_obj = getInput(timestampID, count);
      const TimeStamp& bt = object_cast<TimeStamp>(t_obj);
      fout << bt.time.tv_sec << "." << bt.time.tv_usec << " ";
    }



    Buffer &audioBuffer = *(outputs[audioID].buffer);
	RCPtr<Matrix<float> > outputp(new Matrix<float>(num_channels,length)); // 出力バッファ
	audioBuffer[count] = outputp;
	Matrix<float>& output = *outputp;
        
	int readsize;
        
	/////////////////////////////////////////////
	// 最初のループ (count == 0)
	/////////////////////////////////////////////
	if (count == 0) {
	  // ここで接続を行う
	  cerr << "binding..." << endl;
	  if(bind(c_sock, (struct sockaddr *)&addr, sizeof(addr)) != 0){
		perror("binding error");
		exit(-1);
	  }
	  cerr << "listening..." << endl;
	  if (listen(c_sock, 1) != 0){
		perror("listen error");
		exit(-1);
	  }
	  
	  cerr << "accepting..." << endl;
	  socklen_t len = sizeof(client);
	  sock = accept(c_sock, (struct sockaddr *)&client, &len);
	  if (sock < 0){
		perror("accept error");
		exit(-1);
	  }

	  isConnected = 1;

	  // 出力バッファの調節
	  output.resize(num_channels, length);
          
	  // 指定サイズ読み込み
	  readsize = length * num_channels * sizeof(short int);
	  cerr << "receiving..." << endl;    
	  if (recv(sock, buffer, readsize, 0) == -1)
		eofBuffer[count] = FalseObject;
	  
	  // 出力バッファに書き込み
	  cerr << "writing..." << endl;    
	  for (int r = 0; r < output.nrows(); r++) {
		for (int c = 0; c < output.ncols(); c++) {
		  output(r, c) = buffer[r + c * num_channels];
		}
	  }
	  cerr << "done." << endl;    
	}
	//////////////////////////////////////////////
	// デフォルト (count >= 1)
	//////////////////////////////////////////////
	else {

	  if(!isConnected){
		cerr << "accepting... count:" << count;
		socklen_t len = sizeof(client);
		sock = accept(c_sock, (struct sockaddr *)&client, &len);
		cerr << "sock = " << sock;
		cerr << "  done." << endl;
		isConnected = 1;
	  }
	  
	  // オーバラップあり
	  if (advance < length) {
		Matrix<float>& previous =
		  object_cast<Matrix<float> >(audioBuffer[count - 1]);

		for (int c = 0; c < length - advance; c++) {
		  for (int r = 0; r < output.nrows(); r++) {
			output(r, c) = previous(r, c + advance);
		  }
		}

	  }
	  // なし: そのままコピー
	  else {
		for (int c = 0; c < length - advance; c++) {
		  for (int r = 0; r < output.nrows(); r++) {
			output(r, c) = 0;
		  }
		}
	  }

	  // 指定サイズ読み込み
	  cerr << "recving...";
	  readsize = advance * num_channels * sizeof(short int) ;
	  if (recv(sock, buffer, readsize, MSG_WAITALL) == -1)
		eofBuffer[count] = TrueObject;//FalseObject;
	  cerr << "done" << eofBuffer[count] << endl;

	  int sum=0;
	  for(int tempi=0; tempi<advance*num_channels; tempi++){
		sum += buffer[tempi];
		cout << tempi << ":" << buffer[tempi] << ", ";
	  }
	  cout << endl;
	  cout << "sum: " << sum << endl;
	  if(sum == 0){
		isConnected = 0;
		return;
	  }

	  // 出力バッファに書き込み
	  for (int r = 0; r < output.nrows(); r++) {
		for (int c = length - advance; c < output.ncols(); c++) {
		  output(r, c) = buffer[r + (c - length + advance) * num_channels];
		}
	  }

	}
*/
  }

/*
  int readfull(int id, void *buf, unsigned int size) {
	int ret;
	unsigned int stack = 0, diff = size;
	unsigned char *pbuf = (unsigned char *)buf;
        
	while (stack < size) {
	  if ((ret = read(id, &pbuf[stack], diff)) <= 0)
		return stack;
	  stack += ret;
	  diff  -= ret;
	}
        
	return stack;
  }

    ///////////////////////////////////////////////////////////////
    //
    // connect to a socket server
    //
    ///////////////////////////////////////////////////////////////
    int connectToServer() {
	int socket_id;
	
	hostent* connect_host;
	if ((connect_host = gethostbyname(hostname.c_str())) == NULL) {
	    cerr << "cannot find the host." << endl;
	    return -3;
	}
	
	sockaddr_in desthost;
	bzero((char*)&desthost, sizeof(desthost));
	desthost.sin_family = AF_INET;
	desthost.sin_port = htons(port);
	bcopy(connect_host->h_addr, (char*)&desthost.sin_addr, connect_host->h_length);
	
	if ((socket_id = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	    cerr << "cannot get socket." << endl;
	    return -2;
	}
	
	if (connect(socket_id, (sockaddr*)&desthost, sizeof(desthost)) == -1) {
	    cerr << "connection refused." << endl;
	    close(socket_id);
	    return -1;
	}
	
	return socket_id;
    }
*/
  void hex_string_print(std::string print_msg){
    /*
    for(int i=0; i<print_msg.size(); i++){
      int tc = 0xff & print_msg.c_str()[i];
      cout << hex << "0x" << setw(2) << setfill('0') << internal << tc << " ";
    }
    cout << dec << endl;
    */
  }

  IN_ORDER_NODE_SPEEDUP(PublisherWebSocket)
};

//#endif // ENABLE_WEBSOCKET

