/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Source.h>
#include <Map.h>

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h> 
#include <sys/ioctl.h>
#include <errno.h>

#include "HarkParamsObjects.h"

//using namespace std;
using namespace FD;

class ParamsConvert;

DECLARE_NODE(ParamsConvert);
/*Node
 *
 * @name ParamsConvert
 * @category HARK:CN
 * @description HarkParams dynamic reconfigure node. DO NOT PUT THIS NODE MORE THAN TWO IN THE SAME NETWORK FILE (since single process can generate only one dynamic reconfigure node).
 *
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description print debug message of this module in case of true.
 *
 * @output_name OUT_LocalizeMUSIC
 * @output_type Vector<ObjectRef>
 * @output_description HarkParams of LocalizeMUSIC (ParamsLocalizeMUSIC)
 *
 * @output_name OUT_SourceTracker
 * @output_type Vector<ObjectRef>
 * @output_description HarkParams of SourceTracker (ParamsSourceTracker)
 *
 * @output_name OUT_HRLE
 * @output_type Vector<ObjectRef>
 * @output_description HarkParams of HRLE (ParamsHRLE)
 *
 * @output_name OUT_TFNAME
 * @output_type string
 * @output_description HarkParams of AMATRIX (TFNAME) for LocalizeMUSIC
 *
 * @output_name OUT_LocalizeMUSICVector
 * @output_type Vector<int>
 * @output_description HarkParams of LocalizeMUSIC (Vector<int>(5)) for SaveDataStream
 *
 * @output_name OUT_SourceTrackerVector
 * @output_type Vector<float>
 * @output_description HarkParams of SourceTracker (Vector<float>(5)) for SaveDataStream
 *
 * @output_name OUT_HRLEVector
 * @output_type Vector<float>
 * @output_description HarkParams of HRLE (Vector<float>(2)) for SaveDataStream
 *
 * @output_name OUT_PARAMMAP
 * @output_type any
 * @output_description HarkParams of PARAMMAP (Map<std::string, Map<std::string, ObjectRef> >) for General Parameter Format
 *
END*/

class ParamsConvert : public BufferedNode {
  bool enable_debug;
  int in_localizemusicID;
  int in_sourcetrackerID;
  int in_hrleID;
  int in_tfnameID;
  int in_localizemusic_vecID;
  int in_sourcetracker_vecID;
  int in_hrle_vecID;
  int in_parammapID;

  int out_localizemusicID;
  int out_sourcetrackerID;
  int out_hrleID;
  int out_tfnameID;
  int out_localizemusic_vecID;
  int out_sourcetracker_vecID;
  int out_hrle_vecID;
  int out_parammapID;

public:
  ParamsConvert(std::string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params),
      in_localizemusicID(-1), in_sourcetrackerID(-1), in_hrleID(-1), in_tfnameID(-1),
      in_localizemusic_vecID(-1), in_sourcetracker_vecID(-1), in_hrle_vecID(-1),
      in_parammapID(-1)
  {
    out_localizemusicID     = addOutput("OUT_LocalizeMUSIC");
    out_sourcetrackerID     = addOutput("OUT_SourceTracker");
    out_hrleID              = addOutput("OUT_HRLE");
    out_tfnameID            = addOutput("OUT_TFNAME");
    out_localizemusic_vecID = addOutput("OUT_LocalizeMUSICVector");
    out_sourcetracker_vecID = addOutput("OUT_SourceTrackerVector");
    out_hrle_vecID          = addOutput("OUT_HRLEVector");
    out_parammapID          = addOutput("OUT_PARAMMAP");

    enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));

    inOrder = true;  
  }

  ~ParamsConvert()
  {
  }

  virtual void initialize()
  {
    outputs[out_localizemusicID].lookAhead = max(outputs[out_localizemusicID].lookAhead, outputs[out_sourcetrackerID].lookAhead);
    outputs[out_localizemusicID].lookAhead = max(outputs[out_localizemusicID].lookAhead, outputs[out_hrleID].lookAhead);
    outputs[out_localizemusicID].lookAhead = max(outputs[out_localizemusicID].lookAhead, outputs[out_tfnameID].lookAhead);
    outputs[out_localizemusicID].lookAhead = max(outputs[out_localizemusicID].lookAhead, outputs[out_localizemusic_vecID].lookAhead);
    outputs[out_localizemusicID].lookAhead = max(outputs[out_localizemusicID].lookAhead, outputs[out_sourcetracker_vecID].lookAhead);
    outputs[out_localizemusicID].lookAhead = max(outputs[out_localizemusicID].lookAhead, outputs[out_hrle_vecID].lookAhead);
    outputs[out_localizemusicID].lookAhead = max(outputs[out_localizemusicID].lookAhead, outputs[out_parammapID].lookAhead);
    outputs[out_localizemusicID].lookAhead =
    outputs[out_sourcetrackerID].lookAhead =
    outputs[out_hrleID].lookAhead =
    outputs[out_tfnameID].lookAhead =
    outputs[out_localizemusic_vecID].lookAhead =
    outputs[out_sourcetracker_vecID].lookAhead =
    outputs[out_hrle_vecID].lookAhead =
    outputs[out_parammapID].lookAhead =
    1 + outputs[out_localizemusicID].lookAhead;

    outputs[out_localizemusicID].lookBack = max(outputs[out_localizemusicID].lookBack, outputs[out_sourcetrackerID].lookBack);
    outputs[out_localizemusicID].lookBack = max(outputs[out_localizemusicID].lookBack, outputs[out_hrleID].lookBack);
    outputs[out_localizemusicID].lookBack = max(outputs[out_localizemusicID].lookBack, outputs[out_tfnameID].lookBack);
    outputs[out_localizemusicID].lookBack = max(outputs[out_localizemusicID].lookBack, outputs[out_localizemusic_vecID].lookBack);
    outputs[out_localizemusicID].lookBack = max(outputs[out_localizemusicID].lookBack, outputs[out_sourcetracker_vecID].lookBack);
    outputs[out_localizemusicID].lookBack = max(outputs[out_localizemusicID].lookBack, outputs[out_hrle_vecID].lookBack);
    outputs[out_localizemusicID].lookBack = max(outputs[out_localizemusicID].lookBack, outputs[out_parammapID].lookBack);
    outputs[out_localizemusicID].lookBack =
    outputs[out_sourcetrackerID].lookBack =
    outputs[out_hrleID].lookBack =
    outputs[out_tfnameID].lookBack =
    outputs[out_localizemusic_vecID].lookBack =
    outputs[out_sourcetracker_vecID].lookBack =
    outputs[out_hrle_vecID].lookBack =
    outputs[out_parammapID].lookBack =
    1 + outputs[out_localizemusicID].lookBack;

    std::cout << getName() << " initialized..." << std::endl;
    this->BufferedNode::initialize();
  }

  virtual int translateInput(std::string inputName){
    if(inputName == "IN_LocalizeMUSIC"){
      in_localizemusicID = addInput(inputName);
      return in_localizemusicID;
    }
    else if(inputName == "IN_SourceTracker"){
      in_sourcetrackerID = addInput(inputName);
      return in_sourcetrackerID;
    }
    else if(inputName == "IN_HRLE"){
      in_hrleID = addInput(inputName);
      return in_hrleID;
    }
    else if(inputName == "IN_TFNAME"){
      in_tfnameID = addInput(inputName);
      return in_tfnameID;
    }
    else if(inputName == "IN_LocalizeMUSICVector"){
      in_localizemusic_vecID = addInput(inputName);
      return in_localizemusic_vecID;
    }
    else if(inputName == "IN_SourceTrackerVector"){
      in_sourcetracker_vecID = addInput(inputName);
      return in_sourcetracker_vecID;
    }
    else if(inputName == "IN_HRLEVector"){
      in_hrle_vecID = addInput(inputName);
      return in_hrle_vecID;
    }
    else if(inputName == "IN_PARAMMAP"){
      in_parammapID = addInput(inputName);
      return in_parammapID;
    }
    else if(inputName == "INPUT"){
      for(unsigned int i=0; i<inputs.size(); i++){
        if(inputs[i].name == inputName){
          return i;
        }
      }
    }
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    // output settings
    
    Vector<ObjectRef>& out_localizemusic = *new Vector<ObjectRef>;
    (*(outputs[out_localizemusicID].buffer))[count] = &out_localizemusic;

    Vector<ObjectRef>& out_sourcetracker = *new Vector<ObjectRef>;
    (*(outputs[out_sourcetrackerID].buffer))[count] = &out_sourcetracker;

    Vector<ObjectRef>& out_hrle = *new Vector<ObjectRef>;
    (*(outputs[out_hrleID].buffer))[count] = &out_hrle;

    std::ostringstream os;
    //ObjectRef& out_tfname = *new ObjectRef;
    //RCPtr<ObjectRef> out_tfname(new ObjectRef(new String("")));
    //ObjectRef& out_tfname = *ObjectRef(new String(""));
    //(*(outputs[out_tfnameID].buffer))[count] = &out_tfname;

    RCPtr<Vector<int> > out_localizemusic_vec(new Vector<int>());
    (*(outputs[out_localizemusic_vecID].buffer))[count] = out_localizemusic_vec;

    RCPtr<Vector<float> > out_sourcetracker_vec(new Vector<float>());
    (*(outputs[out_sourcetracker_vecID].buffer))[count] = out_sourcetracker_vec;

    RCPtr<Vector<float> > out_hrle_vec(new Vector<float>());
    (*(outputs[out_hrle_vecID].buffer))[count] = out_hrle_vec;

    RCPtr<Map<std::string, ObjectRef> > out_parammap(new Map<std::string, ObjectRef>);
    (*(outputs[out_parammapID].buffer))[count] = out_parammap;

    // input setting
    
    if(in_parammapID == -1){
      if(in_localizemusicID != -1){
        RCPtr<Vector<ObjectRef> > in_localizemusic_ptr = getInput(in_localizemusicID, count);
        const Vector<ObjectRef>& in_localizemusic_vec = *in_localizemusic_ptr;
        if(in_localizemusic_vec.size() != 0){
          RCPtr<ParamsLocalizeMUSIC> in_localizemusic_val = in_localizemusic_vec[0]; // in
          // output ParamsLocalizeMUSIC structure (safe copy, each member variables)
          RCPtr<ParamsLocalizeMUSIC> out_localizemusic_val(new ParamsLocalizeMUSIC()); // out
          out_localizemusic_val->num_source            = (*in_localizemusic_val).num_source;
          out_localizemusic_val->min_deg               = (*in_localizemusic_val).min_deg;
          out_localizemusic_val->max_deg               = (*in_localizemusic_val).max_deg;
          out_localizemusic_val->lower_bound_frequency = (*in_localizemusic_val).lower_bound_frequency;
          out_localizemusic_val->upper_bound_frequency = (*in_localizemusic_val).upper_bound_frequency;
          out_localizemusic.push_back(out_localizemusic_val);
          // output Vector<int> structure
          out_localizemusic_vec->resize(5);
          (*out_localizemusic_vec)[0] = (*in_localizemusic_val).num_source;
          (*out_localizemusic_vec)[1] = (*in_localizemusic_val).min_deg;
          (*out_localizemusic_vec)[2] = (*in_localizemusic_val).max_deg;
          (*out_localizemusic_vec)[3] = (*in_localizemusic_val).lower_bound_frequency;
          (*out_localizemusic_vec)[4] = (*in_localizemusic_val).upper_bound_frequency;
          // output Map<std::string, Map<std::string, ObjectRef> > structure
          RCPtr<Map<std::string, ObjectRef> > out_localizemusic_map(new Map<std::string, ObjectRef>);
          (*out_parammap)["LocalizeMUSIC"] = out_localizemusic_map;
          (*out_localizemusic_map)["NUM_SOURCE"]            = ObjectRef(new Int((*in_localizemusic_val).num_source));
          (*out_localizemusic_map)["MIN_DEG"]               = ObjectRef(new Int((*in_localizemusic_val).min_deg));
          (*out_localizemusic_map)["MAX_DEG"]               = ObjectRef(new Int((*in_localizemusic_val).max_deg));
          (*out_localizemusic_map)["LOWER_BOUND_FREQUENCY"] = ObjectRef(new Int((*in_localizemusic_val).lower_bound_frequency));
          (*out_localizemusic_map)["UPPER_BOUND_FREQUENCY"] = ObjectRef(new Int((*in_localizemusic_val).upper_bound_frequency));
          if(in_tfnameID != -1){
            ObjectRef in_tfname_ptr = getInput(in_tfnameID, count);
            std::string in_tfname_val = object_cast<std::string>(in_tfname_ptr);
            // output string
            ////out_tfname = ObjectRef(new String(in_tfname_val));
            // output Map<std::string, Map<std::string, ObjectRef> > structure
            (*out_localizemusic_map)["A_MATRIX"] = ObjectRef(new String(in_tfname_val));
          }
          else{
            // output string
            ////out_tfname = ObjectRef(new String(""));
          }
        }
      }
      else if(in_localizemusic_vecID != -1){
        RCPtr<Vector<int> > in_localizemusic_ptr = getInput(in_localizemusic_vecID, count);
        const Vector<int>& in_localizemusic_vec = *in_localizemusic_ptr;
        if(in_localizemusic_vec.size() == 5){
          RCPtr<ParamsLocalizeMUSIC> out_localizemusic_val(new ParamsLocalizeMUSIC()); // out
          // output ParamsLocalizeMUSIC structure (safe copy, each member variables)
          out_localizemusic_val->num_source            = (in_localizemusic_vec)[0];
          out_localizemusic_val->min_deg               = (in_localizemusic_vec)[1];
          out_localizemusic_val->max_deg               = (in_localizemusic_vec)[2];
          out_localizemusic_val->lower_bound_frequency = (in_localizemusic_vec)[3];
          out_localizemusic_val->upper_bound_frequency = (in_localizemusic_vec)[4];
          out_localizemusic.push_back(out_localizemusic_val);
          // output Vector<int> structure
          out_localizemusic_vec->resize(5);
          (*out_localizemusic_vec)[0] = (in_localizemusic_vec)[0];
          (*out_localizemusic_vec)[1] = (in_localizemusic_vec)[1];
          (*out_localizemusic_vec)[2] = (in_localizemusic_vec)[2];
          (*out_localizemusic_vec)[3] = (in_localizemusic_vec)[3];
          (*out_localizemusic_vec)[4] = (in_localizemusic_vec)[4];
          // output Map<std::string, Map<std::string, ObjectRef> > structure
          RCPtr<Map<std::string, ObjectRef> > out_localizemusic_map(new Map<std::string, ObjectRef>);
          (*out_parammap)["LocalizeMUSIC"] = out_localizemusic_map;
          (*out_localizemusic_map)["NUM_SOURCE"]            = ObjectRef(new Int((in_localizemusic_vec)[0]));
          (*out_localizemusic_map)["MIN_DEG"]               = ObjectRef(new Int((in_localizemusic_vec)[1]));
          (*out_localizemusic_map)["MAX_DEG"]               = ObjectRef(new Int((in_localizemusic_vec)[2]));
          (*out_localizemusic_map)["LOWER_BOUND_FREQUENCY"] = ObjectRef(new Int((in_localizemusic_vec)[3]));
          (*out_localizemusic_map)["UPPER_BOUND_FREQUENCY"] = ObjectRef(new Int((in_localizemusic_vec)[4]));
          if(in_tfnameID != -1){
            ObjectRef in_tfname_ptr = getInput(in_tfnameID, count);
            std::string in_tfname_val = object_cast<std::string>(in_tfname_ptr);
            // output string
            ////out_tfname = ObjectRef(new String(in_tfname_val));
            // output Map<std::string, Map<std::string, ObjectRef> > structure
            (*out_localizemusic_map)["A_MATRIX"] = ObjectRef(new String(in_tfname_val));
          }
          else{
            // output string
            ////out_tfname = ObjectRef(new String(""));
          }
        }
      }
      if(in_sourcetrackerID != -1){
        RCPtr<Vector<ObjectRef> > in_sourcetracker_ptr = getInput(in_sourcetrackerID, count);
        const Vector<ObjectRef>& in_sourcetracker_vec = *in_sourcetracker_ptr;
        if(in_sourcetracker_vec.size() != 0){
          RCPtr<ParamsSourceTracker> in_sourcetracker_val = in_sourcetracker_vec[0]; // in
          RCPtr<ParamsSourceTracker> out_sourcetracker_val(new ParamsSourceTracker()); // out
          // output ParamsSourceTracker structure (safe copy, each member variable)
          out_sourcetracker_val->thresh               = (*in_sourcetracker_val).thresh;
          out_sourcetracker_val->pause_length         = (*in_sourcetracker_val).pause_length;
          out_sourcetracker_val->compare_mode         = (*in_sourcetracker_val).compare_mode;
          out_sourcetracker_val->min_src_interval     = (*in_sourcetracker_val).min_src_interval;
          out_sourcetracker_val->min_tfindex_interval = (*in_sourcetracker_val).min_tfindex_interval;
          out_sourcetracker.push_back(out_sourcetracker_val);
          // output Vector<float> structure
          out_sourcetracker_vec->resize(5);
          (*out_sourcetracker_vec)[0]   = (*in_sourcetracker_val).thresh;
          (*out_sourcetracker_vec)[1]   = (*in_sourcetracker_val).pause_length;
          if((*in_sourcetracker_val).compare_mode == Source::DEG){
            (*out_sourcetracker_vec)[2] = 0.0f;
          }
          else if((*in_sourcetracker_val).compare_mode == Source::TFINDEX){
            (*out_sourcetracker_vec)[2] = 1.0f;
          }
          else{
            throw new NodeException(NULL, "Unknown COMPARE_MODE. This must be DEG/TFINDEX", __FILE__, __LINE__);
          }
          (*out_sourcetracker_vec)[3]   = (*in_sourcetracker_val).min_src_interval;
          (*out_sourcetracker_vec)[4]   = (*in_sourcetracker_val).min_tfindex_interval;
          // output Map<std::string, Map<std::string, ObjectRef> > structure
          RCPtr<Map<std::string, ObjectRef> > out_sourcetracker_map(new Map<std::string, ObjectRef>);
          (*out_parammap)["SourceTracker"] = out_sourcetracker_map;
          (*out_sourcetracker_map)["THRESH"]               = ObjectRef(new Float((*in_sourcetracker_val).thresh));
          (*out_sourcetracker_map)["PAUSE_LENGTH"]         = ObjectRef(new Float((*in_sourcetracker_val).pause_length));
          if((*in_sourcetracker_val).compare_mode == Source::DEG){
            (*out_sourcetracker_map)["COMPARE_MODE"]       = ObjectRef(new Float(0.0f));
          }
          else if((*in_sourcetracker_val).compare_mode == Source::TFINDEX){
            (*out_sourcetracker_map)["COMPARE_MODE"]       = ObjectRef(new Float(1.0f));
          }
          else{
            throw new NodeException(NULL, "Unknown COMPARE_MODE. This must be DEG/TFINDEX", __FILE__, __LINE__);
          }
          (*out_sourcetracker_map)["MIN_SRC_INTERVAL"]     = ObjectRef(new Float((*in_sourcetracker_val).min_src_interval));
          (*out_sourcetracker_map)["MIN_TFINDEX_INTERVAL"] = ObjectRef(new Int((int)(*in_sourcetracker_val).min_tfindex_interval));
        }
      }
      else if(in_sourcetracker_vecID != -1){
        RCPtr<Vector<float> > in_sourcetracker_ptr = getInput(in_sourcetracker_vecID, count);
        const Vector<float>& in_sourcetracker_vec = *in_sourcetracker_ptr;
        if(in_sourcetracker_vec.size() == 5){
          RCPtr<ParamsSourceTracker> out_sourcetracker_val(new ParamsSourceTracker()); // out
          // output ParamsSourceTracker structure (safe copy, each member variable)
          out_sourcetracker_val->thresh               = (in_sourcetracker_vec)[0];
          out_sourcetracker_val->pause_length         = (in_sourcetracker_vec)[1];
          if((in_sourcetracker_vec)[2] == 0.0f){
            out_sourcetracker_val->compare_mode       = Source::DEG;
          }
          else if((in_sourcetracker_vec)[2]== 1.0f){
            out_sourcetracker_val->compare_mode       = Source::TFINDEX;
          }
          else{
            throw new NodeException(NULL, "Unknown COMPARE_MODE. This must be DEG/TFINDEX", __FILE__, __LINE__);
          }
          out_sourcetracker_val->min_src_interval     = (in_sourcetracker_vec)[3];
          out_sourcetracker_val->min_tfindex_interval = (in_sourcetracker_vec)[4];
          out_sourcetracker.push_back(out_sourcetracker_val);
          // output Vector<float> structure
          out_sourcetracker_vec->resize(5);
          (*out_sourcetracker_vec)[0] = (in_sourcetracker_vec)[0];
          (*out_sourcetracker_vec)[1] = (in_sourcetracker_vec)[1];
          (*out_sourcetracker_vec)[2] = (in_sourcetracker_vec)[2];
          (*out_sourcetracker_vec)[3] = (in_sourcetracker_vec)[3];
          (*out_sourcetracker_vec)[4] = (in_sourcetracker_vec)[4];
          // output Map<std::string, Map<std::string, ObjectRef> > structure
          RCPtr<Map<std::string, ObjectRef> > out_sourcetracker_map(new Map<std::string, ObjectRef>);
          (*out_parammap)["SourceTracker"] = out_sourcetracker_map;
          (*out_sourcetracker_map)["THRESH"]               = ObjectRef(new Float((in_sourcetracker_vec)[0]));
          (*out_sourcetracker_map)["PAUSE_LENGTH"]         = ObjectRef(new Float((in_sourcetracker_vec)[1]));
          if((in_sourcetracker_vec)[2] == 0.0f){
            (*out_sourcetracker_map)["COMPARE_MODE"]       = ObjectRef(new Float(0.0f));
          }
          else if((in_sourcetracker_vec)[2]== 1.0f){
            (*out_sourcetracker_map)["COMPARE_MODE"]       = ObjectRef(new Float(1.0f));
          }
          else{
            throw new NodeException(NULL, "Unknown COMPARE_MODE. This must be DEG/TFINDEX", __FILE__, __LINE__);
          }
          (*out_sourcetracker_map)["MIN_SRC_INTERVAL"]     = ObjectRef(new Float((in_sourcetracker_vec)[3]));
          (*out_sourcetracker_map)["MIN_TFINDEX_INTERVAL"] = ObjectRef(new Int((int)(in_sourcetracker_vec)[4]));
        }
      }
      if(in_hrleID != -1){
        RCPtr<Vector<ObjectRef> > in_hrle_ptr = getInput(in_hrleID, count);
        const Vector<ObjectRef>& in_hrle_vec = *in_hrle_ptr;
        if(in_hrle_vec.size() != 0){
          RCPtr<ParamsHRLE> in_hrle_val = in_hrle_vec[0]; // in
          RCPtr<ParamsHRLE> out_hrle_val(new ParamsHRLE()); // out
          // output ParamsHRLE structure (safe copy, each member variable)
          out_hrle_val->lx            = (*in_hrle_val).lx;
          out_hrle_val->time_constant = (*in_hrle_val).time_constant;
          // output Vector<float>
          out_hrle_vec->resize(2);
          (*out_hrle_vec)[0] = (*in_hrle_val).lx;
          (*out_hrle_vec)[1] = (*in_hrle_val).time_constant;
          // output Map<std::string, Map<std::string, ObjectRef> > structure
          RCPtr<Map<std::string, ObjectRef> > out_hrle_map(new Map<std::string, ObjectRef>);
          (*out_parammap)["HRLE"] = out_hrle_map;
          (*out_hrle_map)["LX"]            = ObjectRef(new Float((*in_hrle_val).lx));
          (*out_hrle_map)["TIME_CONSTANT"] = ObjectRef(new Float((*in_hrle_val).time_constant));
          ////(*out_hrle_map)["DECAY_FACTOR"]  = ObjectRef(new Int((int)(*in_hrle_val).decay_factor));
        }
      }
      else if(in_hrle_vecID != -1){
        RCPtr<Vector<float> > in_hrle_ptr = getInput(in_hrleID, count);
        const Vector<float>& in_hrle_vec = *in_hrle_ptr;
        if(in_hrle_vec.size() == 2){
          RCPtr<ParamsHRLE> out_hrle_val(new ParamsHRLE()); // out
          // output ParamsHRLE structure (safe copy, each member variable)
          out_hrle_val->lx            = (in_hrle_vec)[0];
          out_hrle_val->time_constant = (in_hrle_vec)[1];
          // output Vector<float>
          out_hrle_vec->resize(2);
          (*out_hrle_vec)[0] = (in_hrle_vec)[0];
          (*out_hrle_vec)[1] = (in_hrle_vec)[1];
          // output Map<std::string, Map<std::string, ObjectRef> > structure
          RCPtr<Map<std::string, ObjectRef> > out_hrle_map(new Map<std::string, ObjectRef>);
          (*out_parammap)["HRLE"] = out_hrle_map;
          (*out_hrle_map)["LX"]            = ObjectRef(new Float((in_hrle_vec)[0]));
          (*out_hrle_map)["TIME_CONSTANT"] = ObjectRef(new Float((in_hrle_vec)[1]));
          ////(*out_hrle_map)["DECAY_FACTOR"]  = ObjectRef(new Int((int)(*in_hrle_vec)[2]));
        }
      }
    }
    else{
      ObjectRef in_parammap_ptr = getInput(in_parammapID, count);
      const Map<std::string, ObjectRef>& p_input_map = object_cast<Map<std::string, ObjectRef> >(in_parammap_ptr);
      Map<std::string, ObjectRef>::const_iterator it;
      for(it = p_input_map.begin(); it != p_input_map.end(); ++it){
        if(it->first == "LocalizeMUSIC"){
          const Map<std::string, ObjectRef>& map_contents = object_cast<Map<std::string, ObjectRef> >(it->second);
          Map<std::string, ObjectRef>::const_iterator it_sub;
          for(it_sub = map_contents.begin(); it_sub != map_contents.end(); ++it_sub){
            if(it_sub->first == "NUM_SOURCE"){
              if(typeid(*it_sub->second) == typeid(Int)){
                //num_source = dereference_cast<int>(it_sub->second);
              }
            }
            else if(it_sub->first == "MIN_DEG"){
              if(typeid(*it_sub->second) == typeid(Int)){
                //min_deg = dereference_cast<int>(it_sub->second);
              }
            }
            else if(it_sub->first == "MAX_DEG"){
              if(typeid(*it_sub->second) == typeid(Int)){
                //max_deg = dereference_cast<int>(it_sub->second);
              }
            }
            else if(it_sub->first == "LOWER_BOUND_FREQUENCY"){
              if(typeid(*it_sub->second) == typeid(Int)){
                //lower_bound_frequency = dereference_cast<int>(it_sub->second);
              }
            }
            else if(it_sub->first == "UPPER_BOUND_FREQUENCY"){
              if(typeid(*it_sub->second) == typeid(Int)){
                //upper_bound_frequency = dereference_cast<int>(it_sub->second);
              }
            }
            else if(it_sub->first == "A_MATRIX"){
              if(typeid(*it_sub->second) == typeid(String)){
                //std::string tfname = object_cast<String>(it_sub->second);
              }
            }
          }
        }
        else if(it->first == "SourceTracker"){
          const Map<std::string, ObjectRef>& map_contents = object_cast<Map<std::string, ObjectRef> >(it->second);
          Map<std::string, ObjectRef>::const_iterator it_sub;
          for(it_sub = map_contents.begin(); it_sub != map_contents.end(); ++it_sub){
            if(it_sub->first == "THRESH"){
              if(typeid(*it_sub->second) == typeid(Float)){
                //thresh = dereference_cast<float>(it_sub->second);
              }
            }
            else if(it_sub->first == "PAUSE_LENGTH"){
              if(typeid(*it_sub->second) == typeid(Float)){
                //pause_length = dereference_cast<float>(it_sub->second);
              }
            }
            else if(it_sub->first == "MIN_SRC_INTERVAL"){
              if(typeid(*it_sub->second) == typeid(Float)){
                //min_source_interval = dereference_cast<float>(it_sub->second);
              }
            }
            else if(it_sub->first == "MIN_TFINDEX_INTERVAL"){
              if(typeid(*it_sub->second) == typeid(Int)){
                //min_tfindex_interval = dereference_cast<int>(it_sub->second);
              }
            }
          }
        }
        else if(it->first == "HRLE"){
          const Map<std::string, ObjectRef>& map_contents = object_cast<Map<std::string, ObjectRef> >(it->second);
          Map<std::string, ObjectRef>::const_iterator it_sub;
          for(it_sub = map_contents.begin(); it_sub != map_contents.end(); ++it_sub){
            if(it_sub->first == "LX"){
              if(typeid(*it_sub->second) == typeid(Float)){
                //lx = dereference_cast<float>(it_sub->second);
              }
            }
            else if(it_sub->first == "TIME_CONSTANT"){
              if(typeid(*it_sub->second) == typeid(Float)){
                //time_constant = dereference_cast<float>(it_sub->second);
              }
            }
            else if(it_sub->first == "DECAY_FACTOR"){
              if(typeid(*it_sub->second) == typeid(Int)){
                //decay_factor = dereference_cast<int>(it_sub->second);
                //int advance = dereference_cast<int>(parameters.get("ADVANCE"));
                //int samples = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
                //time_constant = (float)((1000.0 * advance) / (decay_factor * samples));
              }
            }
          }
        }
      }
    }

  }
};

