/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <unistd.h>

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>

#include <flowdesigner/BufferedNode.h>
#include <flowdesigner/Buffer.h>
#include <flowdesigner/Vector.h>

#include <hark/Map.h>
#include <hark/Source.h>

#include <math.h>
#include <assert.h>

#include "HARK_MQTT.hpp"

using namespace FD;
using namespace HARK_MQTT;

class PublishLocalization;

DECLARE_NODE(PublishLocalization);
/*Node
 *
 * @name PublishLocalization
 * @category HARK:CN
 * @description Publish source infomation.
 *
 * @input_name SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @output_name OUTPUT
 * @output_type Vector<ObjectRef>
 * @output_description The same as input.
 *
 * @parameter_name WINDOW_NAME
 * @parameter_type string
 * @parameter_value Source Location
 * @parameter_description Window name of the time-azimuth map [default: Window name]
 *
 * @parameter_name WINDOW_LENGTH
 * @parameter_type int
 * @parameter_value 1000
 * @parameter_description Window length to show at the same time [sample]
 *
 * @parameter_name ANGLE_RANGE
 * @parameter_type object
 * @parameter_value <Vector<int> -180 180>
 * @parameter_description Accept range of the source angle [deg]
 *
 * @parameter_name COORDINATE_SYSTEM
 * @parameter_type string
 * @parameter_value POLAR
 * @parameter_list POLAR:RECTANGULAR:BOTH
 * @parameter_description Coordinate setting. POLAR is azimuth/elevation, RECTANGULAR is x/y/z.
 * 
 *
 * @parameter_name MQTT_HOST
 * @parameter_type string
 * @parameter_value localhost
 * @parameter_description The host name of MQTT broker
 *
 * @parameter_name MQTT_PORT
 * @parameter_type int
 * @parameter_value 1883
 * @parameter_description The port of MQTT broker
 *
 * @parameter_name MQTT_TOPIC
 * @parameter_type string
 * @parameter_value hark/rawOutput/robot
 * @parameter_description The topic name to send the source
 *
 * @parameter_name MQTT_QoS
 * @parameter_type int
 * @parameter_list 0:1:2
 * @parameter_value 2
 * @parameter_description Quality of Service. The message is sent at most once if 0, at least once if 1, and exactly once if 2.
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Debug option. If true, it prints input list and output list.
 *
 END*/

class PublishLocalization : public BufferedNode {
  bool param_b_debug;
  mqtt_publisher *mqttobj;
  int input_sourcesID;
  int outputID;
  std::string window_name;
  std::string plot_type;
  int min_angle, max_angle;
  int param_i_qos;
  std::map<int, float> sources_map;

public:
  PublishLocalization(std::string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    input_sourcesID = addInput("SOURCES");
    outputID = addOutput("OUTPUT");

    int windowLength = dereference_cast<int>(parameters.get("WINDOW_LENGTH"));

    Vector<int> angle_range;
    angle_range = object_cast<Vector<int> >(parameters.get("ANGLE_RANGE"));
    assert(angle_range.size() == 2);
    if(angle_range[0] < angle_range[1]){ min_angle = angle_range[0]; max_angle = angle_range[1]; }
    else{                                min_angle = angle_range[1]; max_angle = angle_range[0]; }

    window_name = object_cast<String>(parameters.get("WINDOW_NAME"));
    plot_type = object_cast<std::string>(parameters.get("COORDINATE_SYSTEM"));

    // MQTT parameters
    std::string param_s_hostname = object_cast<std::string>(parameters.get("MQTT_HOST"));
    std::string param_s_topic    = object_cast<std::string>(parameters.get("MQTT_TOPIC"));
    int param_i_port             = dereference_cast<int>(parameters.get("MQTT_PORT"));
    param_i_qos                  = dereference_cast<int>(parameters.get("MQTT_QoS"));
    
    param_b_debug                = dereference_cast<bool>(parameters.get("DEBUG"));
    
    //interval_count = interval;
    int keepalive = 60;
    if(param_b_debug){
      std::cerr << "HOST  : " << param_s_hostname << std::endl;
      std::cerr << "PORT  : " << param_i_port << std::endl;
      std::cerr << "TOPIC : " << param_s_topic << std::endl;
      std::cerr << "QOS   : " << param_i_qos << std::endl;
    }
    mqttobj = new mqtt_publisher(/*NULL*/"PUBLISH_LOCALIZATION", param_s_topic, param_s_hostname, param_i_port, keepalive, param_b_debug);
    
    //inputsCache[sourcesID].lookBack = windowLength;
    inOrder = true;
  }

  ~PublishLocalization()
  {
    if(param_b_debug){ std::cerr << "Destructing PublisherFLAC" << std::endl; }
    std::cerr << "Publishing remaining message.";
    while(mqttobj->get_life()>0){ /*mqttobj->request_loop();*/ std::cerr << "."; } std::cerr << " finished." << std::endl;
    delete mqttobj;
    if(param_b_debug){ std::cerr << "Destructing Finished" << std::endl; }
  }

  void initialize(void)
  {
    this->BufferedNode::initialize();

    std::cout << "initialize\n";
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    RCPtr<Vector<ObjectRef> > sources_ptr = getInput(input_sourcesID, count);
    const Vector<ObjectRef>& sources = *sources_ptr;
    int i;
    RCPtr<Source> src;

    out[count] = sources_ptr;

    for(i=0; i<sources.size(); i++) {
      src = sources[i];
      if(sources_map.find(src->id) == sources_map.end()){
        std::map<std::string, std::string> source_map;
        std::ostringstream src_ss("");
        float az = (180 / M_PI * atan2f(src->x[1], src->x[0]));
        while(az > 180){ az -= 360; } while(az < -180){ az += 360; }
        if(0!=0){ // basic mode
          src_ss << "{\"dummy\": \"dummy\"";
          src_ss << ", \"id\": "    << src->id
                 << ", \"x_0\": "   << src->x[0]
                 << ", \"x_1\": "   << src->x[1]
                 << ", \"x_2\": "   << src->x[2]
                 << ", \"power\": " << src->power;
          if(src->compare_mode == src->DEG){ src_ss << ", \"compare_mode\": \"DEG\""; }
          else{                              src_ss << ", \"compare_mode\": \"TFINDEX\""; }
          src_ss << ", \"tfindex_0\": "     << src->tfindex[0]
                 << ", \"tfindex_1\": "     << src->tfindex[1]
                 << ", \"tfindex_2\": "     << src->tfindex[2]
                 << ", \"tfindex_max_0\": " << src->tfindex_max[0]
                 << ", \"tfindex_max_1\": " << src->tfindex_max[1]
                 << ", \"tfindex_max_2\": " << src->tfindex_max[2]
                 << ", \"tfindex_min_0\": " << src->tfindex_min[0]
                 << ", \"tfindex_min_1\": " << src->tfindex_min[1]
                 << ", \"tfindex_min_2\": " << src->tfindex_min[2];
          src_ss << ", \"azimuth\": "   << (180 / M_PI * atan2f(src->x[1], src->x[0]))
                 << ", \"elevation\": " << (180 / M_PI * asin(src->x[2]));
          //timeval tv; gettimeofday(&tv, NULL);
          //src_ss << ", \"tv_sev\": "  << tv.tv_sec
          //       << ", \"tv_usec\": " << tv.tv_usec;
          src_ss << "}";
        }
        else if(0==0){ // mimicry mode
          if((az > 20) && (az < 40)){
            src_ss << "{\"name\": \"mimicry\", \"axis\": \"l_ear\", \"value\": \"on\"}";
          }
          else if((az < -20) && (az > -40)){
            src_ss << "{\"name\": \"mimicry\", \"axis\": \"r_ear\", \"value\": \"on\"}";
          }
        }
        sources_map[src->id] = az;
        mqttobj->send_message(src_ss.str(), param_i_qos, /*count=*/0); // count argument was not effective now.
      }
      else{
        //if you send each frame, please write send code.
      }
      //std::cerr << "[DEBUG] Created new sources_map, servers_map" << std::endl;
    }

    for(std::map<int, float>::iterator itsrc = sources_map.begin(); itsrc != sources_map.end();){
      i=sources.size();
      while(i>0){
        src = sources[i-1];
        if(itsrc->first==src->id){ break; } i--;
      }
      if(i==0){
        std::ostringstream src_ss("");
        float az = itsrc->second;
        sources_map.erase(itsrc++);
        if(0==0){ // mimicry mode
          if((az > 20) && (az < 40)){
            src_ss << "{\"name\": \"mimicry\", \"axis\": \"l_ear\", \"value\": \"off\"}";
          }
          else if((az < -20) && (az > -40)){
            src_ss << "{\"name\": \"mimicry\", \"axis\": \"r_ear\", \"value\": \"off\"}";
          }
        }
        mqttobj->send_message(src_ss.str(), param_i_qos, /*coubnt=*/0); // count argument was not effective now.
        //std::string mark="end";  mqttobj->send_message(mark, 2, count);
      }
      else{
        itsrc++;
        //std::string mark="loop"; mqttobj->send_message(mark, 2, count);
      }
    }

/*
    if (count % 100 == 0) {
      if (prms.window != NULL) {
	if (count > prms.start_time + prms.time_length) {
	  prms.start_time += prms.time_length / 2;
	}

	if (prms.pixmap) {
	  for (int time = 0; time <= prms.time_length; time += 100) {
	    snprintf(label, sizeof(label), "%d", prms.start_time + time);
	    gdk_draw_string(prms.pixmap, font, prms.drawingarea->style->black_gc,
			    prms.plot_area.x + prms.plot_area.width * time / prms.time_length,
			    prms.plot_area.y + prms.plot_area.height + 10,
			    label);
	  }
	}

        pub(count,plot_type);
      }
    }
*/

  }

/*
  void pub(int count, string type) {
    map<int, vector<vector<GdkPoint> > > s;
    int angle;

    for (int t = start_time_f; t <= count; t++) {
      RCPtr<Vector<ObjectRef> > sources = getInput(sourcesID, t);
      for (int i = 0; i < sources->size(); i++) {
        RCPtr<Source> src = (*sources)[i];
        if(type == "POLAR"){
          azimuth = (int)(180 / M_PI * atan2(src->x[1], src->x[0]));
          elevation = (int)(180 / M_PI * atan2(src->x[2], sqrt(src->x[0] * src->x[0] + src->x[1] * src->x[1])));
        }
        else if(type == "RECTANGULAR"){
          angle_x = (int)src->x[0];
          angle_y = (int)src->x[1];
          angle_z = (int)src->x[2];
        }
        else if(type == "BOTH"){
          azimuth = (int)(180 / M_PI * atan2(src->x[1], src->x[0]));
          elevation = (int)(180 / M_PI * atan2(src->x[2], sqrt(src->x[0] * src->x[0] + src->x[1] * src->x[1])));
          angle_x = (int)src->x[0];
          angle_y = (int)src->x[1];
          angle_z = (int)src->x[2];
        }
	bool rflag = false;
	if (lines.find(src->id) != lines.end()){
	  if (lines[src->id][lines[src->id].size()-1].size() > 0){
	    GdkPoint prev_pos = lines[src->id][lines[src->id].size()-1][lines[src->id][lines[src->id].size()-1].size()-1];
	    if (abs(prev_pos.y - pos.y) >  (prms.ymax - prms.ymin) / 2){
	      rflag = true;
	    }
	  }
	  if (rflag) {
	    vector<GdkPoint> linetmp(0);
	    lines[src->id].push_back(linetmp);
	    lines[src->id][lines[src->id].size()-1].push_back(pos);
	  } else {	  
	    lines[src->id][lines[src->id].size()-1].push_back(pos);
	  }
	} else {
	  vector<GdkPoint> linetmp(0);
	  lines[src->id].push_back(linetmp);
	  lines[src->id][lines[src->id].size()-1].push_back(pos);	  
	}
      }
    }

    map<int, vector<vector<GdkPoint> > >::iterator it = lines.begin();
    for (; it != lines.end(); ++it) {
      for (int it2 = 0; it2 < it->second.size(); it2++) {
	gdk_draw_lines(prms.pixmap, prms.plot_gc[it->first % 3],&it->second[it2][0], it->second[it2].size());
      }
    }
  }
*/

};
