/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <unistd.h>

#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include <flowdesigner/BufferedNode.h>
#include <flowdesigner/Buffer.h>
#include <flowdesigner/Vector.h>
#include <flowdesigner/Complex.h>
#include <flowdesigner/Matrix.h>

#include <hark/Map.h>
#include <hark/Source.h>

using namespace FD;

class ReplaceMapID;

DECLARE_NODE(ReplaceMapID);
/*Node
 *
 * @name ReplaceMapID
 * @category HARK:CN
 * @description Replace Map ID from Source for FLAC node, etc...
 *
 * @input_name SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source localization results. Each element of the vector is Source type.
 *
 * @input_name MAP_DATA
 * @input_type Map<int,ObjectRef>
 * @input_description Input Map data. Each element of Map is float data type.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The same as the output.
 *
 END*/

class ReplaceMapID : public BufferedNode {
  int sourcesID;
  int mapID;
  int outputID;
public:
  ReplaceMapID(std::string nodeName, ParameterSet params) : BufferedNode(nodeName, params)
  {
    sourcesID = addInput("SOURCES");
    mapID = addInput("MAP_DATA");
    outputID = addOutput("OUTPUT");
    inOrder = true;
  }
  ~ReplaceMapID(){
  }
  void calculate(int output_id, int count, Buffer &out)
  {
    RCPtr<Vector<ObjectRef> > sources_ptr = getInput(sourcesID, count);
    const Vector<ObjectRef>& sources = *sources_ptr;
    RCPtr<Map<int, ObjectRef> > maps_ptr = getInput(mapID, count);
    const Map<int, ObjectRef>& maps = *maps_ptr;

    Buffer &outBuffer = *(outputs[outputID].buffer);
    RCPtr<Map<int, ObjectRef> > omaps_ptr(new Map<int, ObjectRef>);
    outBuffer[count] = omaps_ptr;
    Map<int, ObjectRef>& omaps = *omaps_ptr;

    for(int i=0; i<sources.size(); i++) {
      RCPtr<Source> src = sources[i];
      if(maps.find(src->id) == maps.end()){
        Map<int, ObjectRef>::const_iterator mapdata_cit = maps.begin();
        Vector<float>& tmpvec = object_cast<Vector<float> >(mapdata_cit->second);
        RCPtr<Vector<float> > vec_float(new Vector<float>(tmpvec.size()));
        for(int elem=0; elem<tmpvec.size(); elem++){
          (*vec_float)[elem] = tmpvec[elem];
        }
        omaps[src->id] = vec_float;
      }
      else{
        Map<int, ObjectRef>::const_iterator mapdata_cit = maps.find(src->id);
        Vector<float>& tmpvec = object_cast<Vector<float> >(mapdata_cit->second);
        RCPtr<Vector<float> > vec_float(new Vector<float>(tmpvec.size()));
        for(int elem=0; elem<tmpvec.size(); elem++){
          (*vec_float)[elem] = tmpvec[elem];
        }
        omaps[src->id] = vec_float;
      }
    }
  }
  IN_ORDER_NODE_SPEEDUP(ReplaceMapID)
};


