/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

//#ifdef ENABLE_MQTT

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <string.h>
#include <vector>
#include <deque>
#include <queue>
#include <map>
#include <time.h>
#include <unistd.h>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>

//#include <mosquittopp.h>
#undef PUBLISHER // bad work, will fix later.
#include "HARK_MQTT.hpp"
#include "HARK_FLAC.hpp"
#include "HARK_UUID.hpp"

using namespace FD;
using namespace HARK_MQTT;
using namespace HARKSF;
using namespace HARK_FLAC;

class SubscriberFLAC;

DECLARE_NODE(SubscriberFLAC);
/*Node
 *
 * @name SubscriberFLAC
 * @category HARK:FLAC
 * @description Subscribe FLAC data via MQTT
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description The same as the output.
 * 
 * @output_name CONDITION
 * @output_type bool
 * @output_description Condition.
 *
 * @parameter_name ENABLE_SAVE_SUBSCRIBED
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Debug option. If true, output subscribed flac stream to files.
 *
 * @parameter_name SAVE_SUBSCRIBED_FILENAME_PREFIX
 * @parameter_type string
 * @parameter_value subscribed_
 * @parameter_description Duplicate from subscrived flac stream per each Source ID to files.[default: subscribed_]
 *
 * @parameter_name ENABLE_SAVE_DECODED
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Debug option. If true, output decoded wav files.
 *
 * @parameter_name SAVE_DECODED_FILENAME_PREFIX
 * @parameter_type string
 * @parameter_value decoded_
 * @parameter_description Duplicate from decoded flac stream per each Source ID to files.[default: decoded_]
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The frame length of each channel (in samples)[default: 512]
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length between adjacent frames (in samples)[default: 160]
 * 
 * @parameter_name MQTT_HOST
 * @parameter_type string
 * @parameter_value localhost
 * @parameter_description The host name of MQTT broker
 *
 * @parameter_name MQTT_PORT
 * @parameter_type int
 * @parameter_value 1883
 * @parameter_description The port of MQTT broker
 *
 * @parameter_name MQTT_TOPIC
 * @parameter_type string
 * @parameter_value hark/data/flac
 * @parameter_description The topic name to send the source
 *
 * @parameter_name MQTT_USERNAME
 * @parameter_type string
 * @parameter_value 
 * @parameter_description The username name to auth MQTT
 *
 * @parameter_name MQTT_PASSWORD
 * @parameter_type string
 * @parameter_value 
 * @parameter_description The password to auth MQTT
 *
 * @parameter_name MQTT_CAFILE
 * @parameter_type string
 * @parameter_value 
 * @parameter_description The ca file to SSL/TSL of MQTT
 *
 * @parameter_name MQTT_TIMEOUT
 * @parameter_type int
 * @parameter_value 60
 * @parameter_description The time out between MQTT messages
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Debug option. If true, it prints input list and output list.
 * 
 END*/


class SubscriberFLAC : public BufferedNode {
  int outputID;
  int conditionID;
  bool param_b_enable_save_subscribed;
  std::string param_s_subscribed_filename_prefix;
  bool param_b_enable_save_decoded;
  std::string param_s_decoded_filename_prefix;
  int param_i_length;
  int param_i_advance;
  int param_i_keepalive;
  int param_i_timeout;
  bool isDebug;

  mqtt_subscriber *mqttobj;
  RequestServer::DecodeServer *decoder_p;
  int recv_size;
  int decode_size;
  bool recv_completed;
  bool exit_flag;
  time_t last_recieved;
  time_t started;
  mqtt_publisher *state_publisher;

  int total_channels;
  std::deque<std::vector<float> > frame_buffer;

public:
  SubscriberFLAC(std::string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    outputID = addOutput("OUTPUT");
    conditionID = addOutput("CONDITION");

    param_b_enable_save_subscribed     = dereference_cast<bool>(parameters.get("ENABLE_SAVE_SUBSCRIBED"));
    param_s_subscribed_filename_prefix = object_cast<std::string>(parameters.get("SAVE_SUBSCRIBED_FILENAME_PREFIX"));
    param_b_enable_save_decoded        = dereference_cast<bool>(parameters.get("ENABLE_SAVE_DECODED"));
    param_s_decoded_filename_prefix    = object_cast<std::string>(parameters.get("SAVE_DECODED_FILENAME_PREFIX"));
    param_i_length                     = dereference_cast<int>(parameters.get("LENGTH"));
    param_i_advance                    = dereference_cast<int>(parameters.get("ADVANCE"));

    // MQTT parameters
    std::string param_s_hostname = object_cast<std::string>(parameters.get("MQTT_HOST"));
    std::string param_s_topic    = object_cast<std::string>(parameters.get("MQTT_TOPIC"));
    int param_i_port             = dereference_cast<int>(parameters.get("MQTT_PORT"));
//    std::string param_s_username = object_cast<std::string>(parameters.get("MQTT_USERNAME"));
//    std::string param_s_password = object_cast<std::string>(parameters.get("MQTT_PASSWORD"));
//    std::string param_s_cafile = object_cast<std::string>(parameters.get("MQTT_CAFILE"));
    std::string param_s_username, param_s_password, param_s_cafile;
    if(parameters.exist("MQTT_USERNAME")){
        param_s_username = object_cast<std::string>(parameters.get("MQTT_USERNAME"));
    }
    if(parameters.exist("MQTT_PASSWORD")){
        param_s_password = object_cast<std::string>(parameters.get("MQTT_PASSWORD"));
    }
    if(parameters.exist("MQTT_CAFILE")){
        param_s_cafile = object_cast<std::string>(parameters.get("MQTT_CAFILE"));
    }

    isDebug                      = dereference_cast<bool>(parameters.get("DEBUG"));
    param_i_keepalive            = 60;
    param_i_timeout              = dereference_cast<int>(parameters.get("MQTT_TIMEOUT"));

    std::cerr << param_s_username << " : " << param_s_password << std::endl;

    HARK_UUID::uuid uuid;
    std::string unique_str = uuid.v4();

    //mosqpp::lib_init(); // no need, included HARK_MQTT::mqtt_subscriber constructor
    //mqttobj = new mqtt_subscriber(/*NULL*/"SUBSCRIBER", param_s_topic, param_s_hostname, param_i_port,
    //    param_s_username, param_s_password, param_s_cafile, param_i_keepalive, true, isDebug);
    std::string subscriber_id("HARK|SubscriberFLAC|" + unique_str);
    mqttobj = new mqtt_subscriber(subscriber_id.c_str(), param_s_topic, param_s_hostname, param_i_port,
        param_s_username, param_s_password, param_s_cafile, param_i_keepalive, true, isDebug);
    if(!mqttobj->is_connected()) throw (new NodeException(NULL,"MQTT connection failed.",__FILE__,__LINE__));

    std::string publisher_id("HARK|SubscriberFLAC|StatePublisher|" + unique_str);
    std::string param_s_state_topic = param_s_topic + "/state";
    //state_publisher= new mqtt_publisher(/*NULL*/"PUBLISHER", param_s_state_topic, param_s_hostname, param_i_port,
    //    param_s_username, param_s_password, param_s_cafile, param_i_keepalive, true, isDebug);
    state_publisher= new mqtt_publisher(publisher_id.c_str(), param_s_state_topic, param_s_hostname, param_i_port,
        param_s_username, param_s_password, param_s_cafile, param_i_keepalive, true, isDebug);

    inOrder = true;
    recv_size = 0;
    decode_size = 0;
    recv_completed = false;
    decoder_p = new RequestServer::DecodeServer();
    decoder_p->start();
    exit_flag = false;
    last_recieved = 0;
    started = time(NULL);

    total_channels = 0;
  };

  virtual void initialize()
  {
    outputs[outputID].lookAhead = outputs[conditionID].lookAhead
      = 1 + std::max(outputs[outputID].lookAhead, outputs[conditionID].lookAhead);
    outputs[outputID].lookBack = outputs[conditionID].lookBack
      = 1 + std::max(outputs[outputID].lookBack, outputs[conditionID].lookBack);

    this->BufferedNode::initialize();
  }

  ~SubscriberFLAC()
  {
    if(isDebug){ std::cerr << "SubscriberFLAC::Destructor"  << std::endl; }
    std::cerr << "COMP decoded size:" << std::dec << decode_size << std::endl;

    //mosqpp::lib_cleanup(); // no need, included HARK_MQTT::mqtt_subscriber destructor
    delete mqttobj;

    delete state_publisher;
    decoder_p->stop();
    delete decoder_p;
  }

  void saveDecodedData(std::vector<float> pcm_vec, bool isLast)
  {
    // not supported  
    return;

    if(!param_b_enable_save_decoded) return;

    std::ostringstream filename_str;
    filename_str << param_s_decoded_filename_prefix << ".wav";

    std::ofstream ofs(filename_str.str(), std::ios::out|std::ios::binary|std::ios::app);
    int bits_per_sample  = decoder_p->get_bits_per_sample(); //ex) 8, 16, 24 [bits];
    int bytes_per_sample = bits_per_sample / 8;
    int channels         = decoder_p->get_channels(); //ex) 1, 2, ... , 8 [channels];
    int sampling_rate    = decoder_p->get_sampling_rate(); //ex) 16000, 48000 [Hz];
    int wav_format       = 0x0001;
    if(ofs){
      std::vector<char> raw_pcm_vec(pcm_vec.size()*bytes_per_sample);
      int value;
      int i, j, k;
      for(i=0, j=0; i<pcm_vec.size(); i++, j+=bytes_per_sample){
        value = pcm_vec[i];
        for(k=0; k<bytes_per_sample; k++){
          raw_pcm_vec[j+k] = (char)((value >> (k*8)) & 0xff);
        }
      }
      // write WAV header
      ofs.write(std::string("RIFF").c_str(), 4);
      value = raw_pcm_vec.size() + 36;
      ofs.write((char*)&value, 4);
      ofs.write(std::string("WAVEfmt ").c_str(), 8);
      value = 16;
      ofs.write((char*)&value, 4);
      value = (channels << 16) | wav_format;
      ofs.write((char*)&value, 4);
      value = sampling_rate;
      ofs.write((char*)&value, 4);
      value = sampling_rate * channels * bytes_per_sample;
      ofs.write((char*)&value, 4);
      value = (bits_per_sample << 16) | (channels * bytes_per_sample);
      ofs.write((char*)&value, 4);
      ofs.write(std::string("data").c_str(), 4);
      value = raw_pcm_vec.size();
      ofs.write((char*)&value, 4);
      // write PCM data
      ofs.write((char*)&raw_pcm_vec[0], raw_pcm_vec.size());
      ofs.close();
    }
  }

  void saveSubscribedData(std::vector<unsigned char> vec)
  {
    if(!param_b_enable_save_subscribed) return;

    std::ostringstream filename_str;
    filename_str << param_s_subscribed_filename_prefix << ".flac";

    std::ofstream ofs(filename_str.str(), std::ios::out|std::ios::binary|std::ios::app);
    if(ofs){
      ofs.write((char*)&vec[0], vec.size());
      ofs.close();
    }
  }

  inline bool starts_with(const std::string& s, const std::string& prefix){
    if(s.size() < prefix.size()){ return false; }
    return std::equal(prefix.begin(), prefix.end(), s.begin());
  }

  inline bool ends_with(const std::string& s, const std::string& suffix){
    if(s.size() < suffix.size()){ return false; }
    return std::equal(s.begin() + s.size() - suffix.size(), s.end(), suffix.begin());
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    while(1){
      // If the session is over (if STOP_STREAM is received), the reception process is skipped.
      if(!recv_completed){
        if(!mqttobj->is_connected()){
          throw (new NodeException(NULL,"MQTT connection failed.",__FILE__,__LINE__));
        }
        // The check cycle to start reception is 100ms, and it will time out
        // if there is no reception within the period specified by MQTT_TIMEOUT parameter
        // from the start of the process.
        if(!recv_size){
          if(param_i_timeout < time(NULL)-started){
            throw (new NodeException(NULL,"MQTT no data available.",__FILE__,__LINE__));
          }

          usleep(100 * 1000);
          state_publisher->send_message("", "READY", 3, 2, 0);
        }

        std::vector<unsigned char> tmpvec;
        do{
          std::string received_data;
          received_data = mqttobj->receive_message();
          if(isDebug){
            std::cerr << "receive size:" << std::dec << received_data.size() << std::endl;
          }
          if(received_data.size() == 0){ continue; }
          // If STOP_STREAM has been received, the remaining FLAC stream is passed to
          // the FLAC decoder along with the end notification of the data stream.
          if(ends_with(received_data, "STOP_STREAM")){
            std::cerr << "STOP_STREAM message receved:: received data size:" << std::dec << recv_size << std::endl;

            if(received_data.size() >= 11){
              received_data.resize(received_data.size() - 11);
            }
            recv_size += received_data.size();
            tmpvec.assign(received_data.begin(), received_data.end());

            decoder_p->requst_decode(tmpvec, tmpvec.size(), true);
            saveSubscribedData(tmpvec);
            tmpvec.clear();

            recv_completed = true;
            //decoder_p->stop(); // I decided to stop the decoder instance in destructor.
            last_recieved = 0; // 0 is substituted to simplify the judgment of 'if' syntax.

            // No data has been sent but a STOP_STREAM has been received.
            if(!recv_size){
              throw (new NodeException(NULL,"MQTT no data available.",__FILE__,__LINE__));
            }
            break;
          }

          recv_size += received_data.size();
          tmpvec.assign(received_data.begin(), received_data.end());

          decoder_p->requst_decode(tmpvec, tmpvec.size(), false);
          saveSubscribedData(tmpvec);
          tmpvec.clear();

          last_recieved = time(NULL);
        } while(mqttobj->has_data());

        // If the MQTT session has not ended, TIMEOUT will be confirmed.
        // It is judged as TIMEOUT if there is no data has been received for more than
        // the period specified by the MQTT_TIMEOUT parameter since the last data received.
        if(last_recieved && param_i_timeout < time(NULL)-last_recieved){
          std::cerr << "TIMEOUT:: reveived data size:" << std::dec << recv_size << std::endl;
          recv_completed = true;
          //decoder_p->stop(); // I decided to stop the decoder instance in destructor.
          last_recieved = 0; // 0 is substituted to simplify the judgment of 'if' syntax.

          if(!recv_size){
            throw (new NodeException(NULL,"MQTT no data available.",__FILE__,__LINE__));
          }
        }
      }

      if(!total_channels){
        total_channels = decoder_p->get_channels();
        // std::cout << "[DEBUG] total_channels: " << total_channels << std::endl;
        if(!total_channels){
          if(recv_completed){
            throw (new NodeException(NULL,"MQTT no data available.",__FILE__,__LINE__));
          }
          usleep(1000);
          continue;
        }
      }

      int read_size = total_channels * param_i_length;
      int del_size = total_channels * param_i_advance;
      bool isLast = false;

      std::vector<float> pcm_vec;
      do{
        pcm_vec = decoder_p->get_decoded_pcm(read_size, del_size);

        if(pcm_vec.size() == 0){
          if(recv_completed){
            // This code is unnecessary because confirmation of the incomplete FLAC frame
            // is performed on the HARK_FLAC side. It should be ...
            //int retry = 100;
            //do{
            //  pcm_vec = decoder_p->get_decoded_pcm(read_size, del_size);
            //  if(pcm_vec.size() != 0){
            //    pcm_vec.resize(read_size, 0.0f);
            //    frame_buffer.push_back(pcm_vec);
            //  }
            //  if(retry-- <= 0){ break; }
            //  usleep(1000);
            //} while(true);
            isLast = true;
            pcm_vec = decoder_p->get_decoded_pcm(read_size, del_size, isLast);
            if(pcm_vec.size() != 0 && pcm_vec.size() < read_size){
              pcm_vec.resize(read_size, 0.0f);
              frame_buffer.push_back(pcm_vec);
            }
          }
          break;
        }

        frame_buffer.push_back(pcm_vec);
        usleep(1000);
      } while(true);


      if(frame_buffer.size() != 0){
        // std::cout << "[DEBUG] frame_buffer.size(): " << frame_buffer.size() << std::endl;
        pcm_vec = frame_buffer.front();

        // output decoded pcm to OUTPUT
        RCPtr<Matrix<float> > output_matrix(new Matrix<float>(total_channels, param_i_length));
        (*(outputs[outputID].buffer))[count] = output_matrix;
        Matrix<float>& output = *output_matrix;

        decode_size += isLast ? del_size : pcm_vec.size();

        int pos = 0;
        for(int l=0; l<param_i_length; l++){
          for(int c=0; c<total_channels; c++){
            output(c, l) = pcm_vec[pos++];
          }
        }

        saveDecodedData(pcm_vec, isLast);

        frame_buffer.pop_front();

        break;
      }
      else if(recv_completed){
        // output decoded pcm to OUTPUT
        RCPtr<Matrix<float> > output_matrix(new Matrix<float>(total_channels, param_i_length));
        (*(outputs[outputID].buffer))[count] = output_matrix;
        Matrix<float>& output = *output_matrix;

        for(int l=0; l<param_i_length; l++){
          for(int c=0; c<total_channels; c++){
            output(c, l) = 0.0f;
          }
        }
        exit_flag = true;

        break;
      }

      usleep(1000); // Process is too fast, cannot debugging.
    }
    // std::cout << "[DEBUG] count: " << count << std::endl;

    Buffer &output_conditionBuffer = *(outputs[conditionID].buffer);
    output_conditionBuffer[count] = (exit_flag ? FalseObject : TrueObject);
  }
  IN_ORDER_NODE_SPEEDUP(SubscriberFLAC);
};

//#endif // ENABLE_MQTT

