/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

//#ifdef ENABLE_MQTT

#include <iostream>
#include <sstream>
#include <string>
#include <string.h>
#include <vector>
#include <queue>
#include <map>

#include <unistd.h>

#include <flowdesigner/BufferedNode.h>
#include <flowdesigner/Buffer.h>
#include <flowdesigner/Vector.h>
#include <flowdesigner/Complex.h>
#include <flowdesigner/Matrix.h>

#include <hark/Map.h>
#include <hark/Source.h>

//#include <mosquittopp.h>
#undef PUBLISHER // bad work, will fix later.
#include "HARK_MQTT.hpp"
#include "HARK_FLAC.hpp"

using namespace FD;
using namespace HARK_MQTT;
//using namespace HARK_FLAC;
using namespace HARKSF::HARK_FLAC; // bad work, will fix later.

class SubscriberFLAC;

DECLARE_NODE(SubscriberFLAC);
/*Node
 *
 * @name SubscriberFLAC
 * @category HARK:FLAC
 * @description Subscribe FLAC data via MQTT
 *
 * @output_name OUT_SOURCES
 * @output_type Vector<ObjectRef>
 * @output_description The same as the output.
 * 
 * @output_name OUT_PCM
 * @output_type any
 * @output_description The same as the output.
 * 
 * @output_name CONDITION
 * @output_type bool
 * @output_description Condition.
 *
 * @parameter_name ENABLE_SAVE_SUBSCRIBED
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Debug option. If true, output subscribed flac stream to files.
 *
 * @parameter_name SAVE_SUBSCRIBED_FILENAME_PREFIX
 * @parameter_type string
 * @parameter_value subscribed_
 * @parameter_description Duplicate from subscrived flac stream per each Source ID to files.[default: subscribed_]
 *
 * @parameter_name ENABLE_SAVE_DECODED
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Debug option. If true, output decoded wav files.
 *
 * @parameter_name SAVE_DECODED_FILENAME_PREFIX
 * @parameter_type string
 * @parameter_value decoded_
 * @parameter_description Duplicate from decoded flac stream per each Source ID to files.[default: decoded_]
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The frame length of each channel (in samples)[default: 512]
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length between adjacent frames (in samples)[default: 160]
 * 
 * @parameter_name MQTT_HOST
 * @parameter_type string
 * @parameter_value localhost
 * @parameter_description The host name of MQTT broker
 *
 * @parameter_name MQTT_PORT
 * @parameter_type int
 * @parameter_value 1883
 * @parameter_description The port of MQTT broker
 *
 * @parameter_name MQTT_TOPIC
 * @parameter_type string
 * @parameter_value hark/data/flac
 * @parameter_description The topic name to send the source
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Debug option. If true, it prints input list and output list.
 * 
 END*/

class SubscriberFLAC : public BufferedNode {
  bool param_b_enable_save_subscribed;
  bool param_b_enable_save_decoded;
  bool isDebug;
  mqtt_subscriber *mqttobj;
  int output_sourcesID;
  int output_pcmID;
  int output_conditionID;
  int param_i_length;
  int param_i_advance;
  std::string param_s_subscribed_filename_prefix;
  std::string param_s_decoded_filename_prefix;
  Vector<Source> srcbuf;
  int rc;
  
  //  std::map<int, std::queue<float> > decoded_buffer;
  std::map<int, Source> sources_map;
  std::map<int, std::queue<std::vector<float> > > pcm_map;
  std::map<int, RequestServer::DecodeServer *> servers_map;
public:
  SubscriberFLAC(std::string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    output_sourcesID = addOutput("OUT_SOURCES");
    output_pcmID = addOutput("OUT_PCM");
    output_conditionID = addOutput("CONDITION");

    param_b_enable_save_subscribed     = dereference_cast<bool>(parameters.get("ENABLE_SAVE_SUBSCRIBED"));
    param_s_subscribed_filename_prefix = object_cast<std::string>(parameters.get("SAVE_SUBSCRIBED_FILENAME_PREFIX"));
    param_b_enable_save_decoded        = dereference_cast<bool>(parameters.get("ENABLE_SAVE_DECODED"));
    param_s_decoded_filename_prefix    = object_cast<std::string>(parameters.get("SAVE_DECODED_FILENAME_PREFIX"));
    param_i_length                     = dereference_cast<int>(parameters.get("LENGTH"));
    param_i_advance                    = dereference_cast<int>(parameters.get("ADVANCE"));

    // MQTT parameters
    std::string param_s_hostname = object_cast<std::string>(parameters.get("MQTT_HOST"));
    std::string param_s_topic    = object_cast<std::string>(parameters.get("MQTT_TOPIC"));
    int param_i_port             = dereference_cast<int>(parameters.get("MQTT_PORT"));
    
    isDebug         = dereference_cast<bool>(parameters.get("DEBUG"));
    int param_i_keepalive        = 60;

    //mosqpp::lib_init(); // no need, included HARK_MQTT::mqtt_subscriber constructor
    mqttobj = new mqtt_subscriber(/*NULL*/"SUBSCRIBER", param_s_topic, param_s_hostname, param_i_port, param_i_keepalive, isDebug);
    inOrder = true;
  };
  ~SubscriberFLAC(){
    //mosqpp::lib_cleanup(); // no need, included HARK_MQTT::mqtt_subscriber destructor
    delete mqttobj;
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    Buffer &output_sourcesBuffer = *(outputs[output_sourcesID].buffer);
    Vector<ObjectRef>& output_sources_p = *new Vector<ObjectRef>;
    output_sourcesBuffer[count] = &output_sources_p;

    Buffer &output_pcmBuffer = *(outputs[output_pcmID].buffer);
    Map<int, ObjectRef>& output_pcm_p = *new Map<int, ObjectRef>;
//    RCPtr<Map<int, ObjectRef> > output_pcm_map_p(new Map<int, ObjectRef>);
//    Map<int, ObjectRef>& output_pcm_p = *output_pcm_map_p;
    output_pcmBuffer[count] = &output_pcm_p;

//    RCPtr<Map<int, ObjectRef> > output_pcm_map_p(new Map<int, ObjectRef>);
//    output_pcmBuffer[count] = &output_pcm_p;
//    Map<int, ObjectRef>& output_pcm_p = *output_pcm_map_p;

/*
    RCPtr<Source> src(new Source);
    std::istringstream src_ss("");
    src_ss.str("0");
    src_ss >> src->id;
    src_ss.str("0.0f");
    src_ss >> src->x[0];
    src_ss.str("0.0f");
    src_ss >> src->x[1];
    src_ss.str("0.0f");
    src_ss >> src->x[2];
    src_ss.str("0.0f");
    src_ss >> src->power;
    if("DEG" == "DEG"){ src->compare_mode = src->DEG; }
    else{               src->compare_mode = src->TFINDEX; }
    src_ss.str("0");
    src_ss >> src->tfindex[0];
    src_ss.str("0");
    src_ss >> src->tfindex[1];
    src_ss.str("0");
    src_ss >> src->tfindex[2];
    src_ss.str("0");
    src_ss >> src->tfindex_max[0];
    src_ss.str("0");
    src_ss >> src->tfindex_max[1];
    src_ss.str("0");
    src_ss >> src->tfindex_max[2];
    src_ss.str("0");
    src_ss >> src->tfindex_min[0];
    src_ss.str("0");
    src_ss >> src->tfindex_min[1];
    src_ss.str("0");
    src_ss >> src->tfindex_min[2];
    output_sources_p.push_back(src);
*/

    Buffer &output_conditionBuffer = *(outputs[output_conditionID].buffer);
    output_conditionBuffer[count] = TrueObject;

    //[NOTE] using none blocking algorithm, no need following like receive loop.
    //std::string received_data;
    //bool receiving_flag = true;
    //while(receiving_flag){
    //  received_data = mqttobj->receive_message();
    //  if(received_data != ""){ receiving_flag = false; }
    //}

    std::string received_data;
    received_data = mqttobj->receive_message();
    if(received_data != ""){
      std::vector<unsigned char> tmpvec(received_data.begin(), received_data.end());
      RequestServer::DecodeServer *decoder_p = new RequestServer::DecodeServer();
      decoder_p->start();
      decoder_p->requst_decode(tmpvec, tmpvec.size(), true);
      decoder_p->stop();

      std::vector<float> pcm_vec;
      do{ pcm_vec = decoder_p->get_decoded_pcm(); } while(pcm_vec.size()==0);
      //[DEBUG code] dump received data
      //std::cerr << "Received FLAC Data:" << std::endl;
      //for(std::vector<float>::iterator it=pcm_vec.begin(); it<pcm_vec.end(); it++){
      //  std::cerr << std::hex << std::setw(2) << ((int) *it & 0xff) << " ";
      //}
      //std::cerr << std::dec << std::endl;

      std::map<std::string, std::string> meta;
      size_t meta_size = decoder_p->get_metadata(meta);
      std::map<std::string, std::string>::iterator it=meta.begin();
      std::istringstream src_ss("");
      //Source *src = new Source;
      Source src_x;
      Source *src = &src_x;
      float f_val;
      int i_val;
      while(it != meta.end()){
        //[DEBUG code]
        //std::cerr << "key: " << it->first << " value: " << it->second << std::endl;
/*
        src_ss.str(it->second);
        if(     (it->first).compare("id")==0){            src_ss >> src->id; }
        else if((it->first).compare("x_0")==0){           src_ss >> src->x[0]; }
        else if((it->first).compare("x_1")==0){           src_ss >> src->x[1]; }
        else if((it->first).compare("x_2")==0){           src_ss >> src->x[2]; }
        else if((it->first).compare("power")==0){         src_ss >> src->power; }
        else if((it->first).compare("compare_mode")==0){
          if((it->second).compare("DEG")==0){             src->compare_mode = src->DEG; }
          else{                                           src->compare_mode = src->TFINDEX; }
        }
        else if((it->first).compare("tfindex_0")==0){     src_ss >> src->tfindex[0]; }
        else if((it->first).compare("tfindex_1")==0){     src_ss >> src->tfindex[1]; }
        else if((it->first).compare("tfindex_2")==0){     src_ss >> src->tfindex[2]; }
        else if((it->first).compare("tfindex_max_0")==0){ src_ss >> src->tfindex_max[0]; }
        else if((it->first).compare("tfindex_max_1")==0){ src_ss >> src->tfindex_max[1]; }
        else if((it->first).compare("tfindex_max_2")==0){ src_ss >> src->tfindex_max[2]; }
        else if((it->first).compare("tfindex_min_0")==0){ src_ss >> src->tfindex_min[0]; }
        else if((it->first).compare("tfindex_min_1")==0){ src_ss >> src->tfindex_min[1]; }
        else if((it->first).compare("tfindex_min_2")==0){ src_ss >> src->tfindex_min[2]; }
*/
        if(     (it->first).compare("id")==0){            src->id    = std::atoi((it->second).c_str()); }
        else if((it->first).compare("x_0")==0){           src->x[0]  = std::atof((it->second).c_str()); }
        else if((it->first).compare("x_1")==0){           src->x[1]  = std::atof((it->second).c_str()); }
        else if((it->first).compare("x_2")==0){           src->x[2]  = std::atof((it->second).c_str()); }
        else if((it->first).compare("power")==0){         src->power = std::atof((it->second).c_str()); }
        else if((it->first).compare("compare_mode")==0){
          if((it->second).compare("DEG")==0){             src->compare_mode = src->DEG; }
          else{                                           src->compare_mode = src->TFINDEX; }
        }
        else if((it->first).compare("tfindex_0")==0){     src->tfindex[0] = std::atoi((it->second).c_str()); }
        else if((it->first).compare("tfindex_1")==0){     src->tfindex[1] = std::atoi((it->second).c_str()); }
        else if((it->first).compare("tfindex_2")==0){     src->tfindex[2] = std::atoi((it->second).c_str()); }
        else if((it->first).compare("tfindex_max_0")==0){ src->tfindex_max[0] = std::atoi((it->second).c_str()); }
        else if((it->first).compare("tfindex_max_1")==0){ src->tfindex_max[1] = std::atoi((it->second).c_str()); }
        else if((it->first).compare("tfindex_max_2")==0){ src->tfindex_max[2] = std::atoi((it->second).c_str()); }
        else if((it->first).compare("tfindex_min_0")==0){ src->tfindex_min[0] = std::atoi((it->second).c_str()); }
        else if((it->first).compare("tfindex_min_1")==0){ src->tfindex_min[1] = std::atoi((it->second).c_str()); }
        else if((it->first).compare("tfindex_min_2")==0){ src->tfindex_min[2] = std::atoi((it->second).c_str()); }
        it++;
      }
      //[DEBUG code]
      //std::cerr << "id:" << src->id << " x[0]:" << src->x[0] << " x[1]:" << src->x[1] << " x[2]:" << src->x[2] << " power:" << src->power << std::endl;
      //std::cerr << "tfindex[0]:" << src->tfindex[0] << " tfindex[1]:" << src->tfindex[1] << " tfindex[2]:" << src->tfindex[2] << std::endl;
      //std::cerr << "tfindex_max[0]:" << src->tfindex_max[0] << " tfindex_max[1]:" << src->tfindex_max[1] << " tfindex_max[2]:" << src->tfindex_max[2] << std::endl;
      //std::cerr << "tfindex_min[0]:" << src->tfindex_min[0] << " tfindex_min[1]:" << src->tfindex_min[1] << " tfindex_min[2]:" << src->tfindex_min[2] << std::endl;

      std::queue<std::vector<float> > queue_buffer;
      int i, j, k, shift_offset;
      for(i=0; i<pcm_vec.size()-param_i_length; i+=param_i_advance){
        std::vector<float> pcm_tmpvec(param_i_length);
        for(j=i, k=0; k<param_i_length; j++, k++){
          pcm_tmpvec[k] = pcm_vec[j];
        }
        queue_buffer.push(pcm_tmpvec);
      }
      pcm_map[src->id] = queue_buffer;
      //sources_map[src->id] = *src;
      sources_map[src->id] = src_x;
      //servers_map[src->id] = decoder_p; // if you require stream read, this map is need.

      if(param_b_enable_save_subscribed){
        std::ostringstream filename_str;
        std::ofstream ofs;
        filename_str << param_s_subscribed_filename_prefix << src->id << ".flac";
        ofs.open(filename_str.str().c_str(), std::ios::out|std::ios::binary|std::ios::trunc);
        if(ofs){
          ofs.write((char*)&tmpvec[0], tmpvec.size());
          ofs.close();
        }
      }

      if(param_b_enable_save_decoded){
        std::ostringstream filename_str;
        std::ofstream ofs;
        filename_str << param_s_decoded_filename_prefix << src->id << ".wav";
        ofs.open(filename_str.str().c_str(), std::ios::out|std::ios::binary|std::ios::trunc);
        int bits_per_sample  = decoder_p->get_bits_per_sample(); //ex) 8, 16, 24 [bits];
        int bytes_per_sample = bits_per_sample / 8;
        int channels         = decoder_p->get_channels(); //ex) 1, 2, ... , 8 [channels];
        int sampling_rate    = decoder_p->get_sampling_rate(); //ex) 16000, 48000 [Hz];
        int wav_format       = 0x0001;
        if(ofs){
          std::vector<char> raw_pcm_vec(pcm_vec.size()*bytes_per_sample);
          int value;
          for(i=0, j=0; i<pcm_vec.size(); i++, j+=bytes_per_sample){
            value = pcm_vec[i];
            for(k=0; k<bytes_per_sample; k++){
              raw_pcm_vec[j+k] = (char)((value >> (k*8)) & 0xff);
            }
          }
          // write WAV header
          ofs.write(std::string("RIFF").c_str(), 4);
          value = raw_pcm_vec.size() + 36;
          ofs.write((char*)&value, 4);
          ofs.write(std::string("WAVEfmt ").c_str(), 8);
          value = 16;
          ofs.write((char*)&value, 4);
          value = (channels << 16) | wav_format;
          ofs.write((char*)&value, 4);
          value = sampling_rate;
          ofs.write((char*)&value, 4);
          value = sampling_rate * channels * bytes_per_sample;
          ofs.write((char*)&value, 4);
          value = (bits_per_sample << 16) | (channels * bytes_per_sample);
          ofs.write((char*)&value, 4);
          ofs.write(std::string("data").c_str(), 4);
          value = raw_pcm_vec.size();
          ofs.write((char*)&value, 4);
          // write PCM data
          ofs.write((char*)&raw_pcm_vec[0], raw_pcm_vec.size());
          ofs.close();
        }
      }

      //decoder_p->stop();
      delete decoder_p;
    }

    for(std::map<int, Source>::iterator srcit=sources_map.begin(); srcit!=sources_map.end();){
      std::map<int, std::queue<std::vector<float> > >::iterator pcmit = pcm_map.find(srcit->first);
      if(pcmit != pcm_map.end()){ // if not terminate source found
        //RCPtr<Source> src(&(it->second)); // cannot this. batchflow delete automatically per each frame.
        RCPtr<Source> src(new Source);
        src->id             = (srcit->second).id;
        src->x[0]           = (srcit->second).x[0];
        src->x[1]           = (srcit->second).x[1];
        src->x[2]           = (srcit->second).x[2];
        src->power          = (srcit->second).power;
        src->compare_mode   = (srcit->second).compare_mode;
        src->tfindex[0]     = (srcit->second).tfindex[0];
        src->tfindex[1]     = (srcit->second).tfindex[1];
        src->tfindex[2]     = (srcit->second).tfindex[2];
        src->tfindex_max[0] = (srcit->second).tfindex_max[0];
        src->tfindex_max[1] = (srcit->second).tfindex_max[1];
        src->tfindex_max[2] = (srcit->second).tfindex_max[2];
        src->tfindex_min[0] = (srcit->second).tfindex_min[0];
        src->tfindex_min[1] = (srcit->second).tfindex_min[1];
        src->tfindex_min[2] = (srcit->second).tfindex_min[2];
        output_sources_p.push_back(src);
        
        RCPtr<Vector<float> > vec_float(new Vector<float>(param_i_length));
        int i;
        std::vector<float> pcm_tmpvec = (pcmit->second).front();
        for(i=0; i<param_i_length; i++){
          (*vec_float)[i] = pcm_tmpvec[i];
        }
        (pcmit->second).pop();
        if((pcmit->second).empty()){ pcm_map.erase(pcmit); }
        output_pcm_p[srcit->first] = vec_float;
        
        srcit++;
      }
      else{ // if terminate source found
        sources_map.erase(srcit++);
      }
    }

    //[DEBUG code] dump received data
    //std::cerr << "Received FLAC Data:" << std::endl;
    //for(std::string::iterator it=received_data.begin(); it<received_data.end(); it++){
    //  std::cerr << std::hex << std::setw(2) << ((int) *it & 0xff) << " ";
    //}
    //std::cerr << std::dec << std::endl;

    //[NOTE] replace from loop() to loop_start(). I think no problem on subscriber.
    //rc = mqttobj->loop();
    //if(rc){ mqttobj->reconnect(); std::cerr << "recconected." << std::endl; }

    //[DEBUG code]
    //std::cerr << count << std::endl;
    
    if(sources_map.size()==0){
      usleep(1000); // Process is too fast, cannot debugging.
    }
  }
  IN_ORDER_NODE_SPEEDUP(SubscriberFLAC);
};

//#endif // ENABLE_MQTT

